"""Provide common functions and types for the code generation."""
import ast
import inspect
import io
import itertools
import re
import textwrap
from typing import (
    Optional,
    Tuple,
    cast,
    List,
    overload,
    Union,
    Iterator,
    Sequence,
    NoReturn,
    Any,
    Iterable,
    TypeVar,
)

import asttokens
from icontract import require, DBC, ensure


class Rstripped(str):
    """
    Represent a block of text without trailing whitespace.

    The block can be both single-line or multi-line.
    """

    @require(
        lambda block: not block.endswith("\n")
        and not block.endswith(" ")
        and not block.endswith("\t")
    )
    def __new__(cls, block: str) -> "Rstripped":
        return cast(Rstripped, block)


def is_stripped(text: str) -> bool:
    """Check that the ``text`` does not have leading and trailing whitespace."""
    return (
        not text.startswith("\n")
        and not text.startswith(" ")
        and not text.startswith("\t")
    ) and (
        not text.endswith("\n") and not text.endswith(" ") and not text.endswith("\t")
    )


class Stripped(Rstripped):
    """
    Represent a block of text without leading and trailing whitespace.

    The block of text can be both single-line and multi-line.
    """

    @require(lambda block: is_stripped(block))
    def __new__(cls, block: str) -> "Stripped":
        return cast(Stripped, block)


# noinspection RegExpSimplifiable
IDENTIFIER_RE = re.compile(r"[a-zA-Z_][a-zA-Z_0-9]*")


class Identifier(DBC, Stripped):
    """Represent an identifier."""

    @require(lambda value: IDENTIFIER_RE.fullmatch(value))
    def __new__(cls, value: str) -> "Identifier":
        return cast(Identifier, value)


class Error:
    """
    Represent an unexpected input.

    For example, the code of the meta-model can be a valid Python code, but we
    can only process a subset of language constructs.
    """

    def __init__(
        self,
        node: Optional[ast.AST],
        message: str,
        underlying: Optional[List["Error"]] = None,
    ) -> None:
        self.node = node
        self.message = message
        self.underlying = underlying

    def __repr__(self) -> str:
        return (
            f"Error("
            f"node={self.node!r}, "
            f"message={self.message!r}, "
            f"underlying={self.underlying!r})"
        )


class LinenoColumner:
    """Map the source code to line number and column for precise error messages."""

    def __init__(self, atok: asttokens.ASTTokens) -> None:
        positions = []  # type: List[Tuple[int, int]]
        lineno = 1
        column = 0
        for character in atok.get_text(atok.tree):
            if character == "\n":
                column = 1
                lineno += 1
            else:
                column += 1

            positions.append((lineno, column))

        self.atok = atok
        self.positions = positions

    def error_message(self, error: Error) -> str:
        """Generate the error message based on the unexpected observation."""
        prefix = ""
        if error.node is not None:
            start, _ = self.atok.get_text_range(node=error.node)
            lineno, column = self.positions[start]

            prefix = f"At line {lineno} and column {column}: "

        if error.underlying is None or len(error.underlying) == 0:
            return f"{prefix}{error.message}"
        else:
            writer = io.StringIO()
            writer.write(f"{prefix}{error.message}\n")
            for i, underlying_error in enumerate(error.underlying):
                if i > 0:
                    writer.write("\n")
                indented = textwrap.indent(self.error_message(underlying_error), "  ")
                writer.write(indented)

            return writer.getvalue()


class Lines(DBC):
    """Represent a sequence of text lines."""

    # fmt: off
    @require(
        lambda lines:
        all('\n' not in line and '\r' not in line for line in lines)
    )
    # fmt: on
    def __new__(cls, lines: Sequence[str]) -> "Lines":
        r"""
        Ensure the properties on the ``lines``.

        Please make sure that you transfer the "ownership" immediately to Lines
        and don't modify the original list of strings any more:

        .. code-block:: python

            ##
            # OK
            ##

            lines = Lines(some_text.splitlines())

            ##
            # Not OK
            ##

            some_lines = some_text.splitlines()
            lines = Lines(some_lines)
            # ... do something assuming ``lines`` is immutable ...

            some_lines[0] = "This will break \n your logic"
            # ERROR! lines[0] now contains a new-line which is unexpected!

        """
        return cast(Lines, lines)

    # pylint: disable=function-redefined

    @overload
    def __getitem__(self, index: int) -> str:
        """Get the item at the given integer index."""
        raise NotImplementedError("Only for type annotations")

    @overload
    def __getitem__(self, index: slice) -> "Lines":
        """Get the slice of the lines."""
        raise NotImplementedError("Only for type annotations")

    def __getitem__(self, index: Union[int, slice]) -> Union[str, "Lines"]:
        """Get the line(s) at the given index."""
        raise NotImplementedError("Only for type annotations")

    def __iter__(self) -> Iterator[str]:
        """Iterate over the lines."""
        raise NotImplementedError("Only for type annotations")

    def __len__(self) -> int:
        """Return the number of the lines."""
        raise NotImplementedError("Only for type annotations")

    def __add__(self, other: "Lines") -> "Lines":
        """Concatenate two list of lines."""
        raise NotImplementedError("Only for type annotations")


def assert_never(value: NoReturn) -> NoReturn:
    """
    Signal to mypy to perform an exhaustive matching.

    Please see the following page for more details:
    https://hakibenita.com/python-mypy-exhaustive-checking
    """
    assert False, f"Unhandled value: {value} ({type(value).__name__})"


def indent_but_first_line(text: str, indention: str) -> str:
    """
    Indent all but the first of the given ``text`` by ``indention``.

    For example, this helps you insert indented blocks into formatted string literals.
    """
    indented_lines = []  # type: List[str]
    for i, line in enumerate(text.splitlines()):
        if i == 0:
            indented_lines.append(line)
        else:
            if len(line) > 0:
                indented_lines.append(indention + line)
            else:
                indented_lines.append(line)

    return "\n".join(indented_lines)


def assert_union_of_descendants_exhaustive(union: Any, base_class: Any) -> None:
    """
    Check that the ``union`` covers all the concrete subclasses of ``base_class``.

    Make sure you put the assertion at the end of the module where no new classes are
    defined.

    See also for more details: https://hakibenita.com/python-mypy-exhaustive-checking
    """
    if inspect.isclass(union):
        union_map = {id(union): union}
    elif hasattr(union, "__args__"):
        union_map = {id(cls): cls for cls in union.__args__}
    else:
        raise NotImplementedError(f"We do not know how to handle the union: {union}")

    # We have to recursively figure out the subclasses.
    concrete_subclasses = []  # type: List[Any]

    stack = base_class.__subclasses__()  # type: List[Any]

    while len(stack) > 0:
        sub_cls = stack.pop()
        if not inspect.isabstract(sub_cls):
            concrete_subclasses.append(sub_cls)

        stack.extend(sub_cls.__subclasses__())

    subclass_map = {id(sub_cls): sub_cls for sub_cls in concrete_subclasses}

    union_set = set(union_map.keys())
    subclass_set = set(subclass_map.keys())

    if union_set != subclass_set:
        union_diff = union_set.difference(subclass_set)
        union_diff_names = [union_map[cls_id].__name__ for cls_id in union_diff]

        subclass_diff = subclass_set.difference(union_set)
        subclass_diff_names = [
            subclass_map[cls_id].__name__ for cls_id in subclass_diff
        ]

        if len(union_diff_names) == 0 and len(subclass_diff_names) > 0:
            raise AssertionError(
                f"The following concrete subclasses of {base_class.__name__!r} were "
                f"not listed in the union: {subclass_diff_names}"
            )

        elif len(union_diff_names) > 0 and len(subclass_diff_names) == 0:
            raise AssertionError(
                f"The following classes were listed in the union, "
                f"but they are not sub-classes "
                f"of {base_class.__name__!r}: {union_diff_names}"
            )
        else:
            raise AssertionError(
                f"The following classes were listed in the union, "
                f"but they are not sub-classes "
                f"of {base_class.__name__!r}: {union_diff_names}.\n\n"
                f"The following concrete sub-classes of {base_class.__name__!r} were "
                f"not listed in the union: {subclass_diff_names}"
            )


def assert_union_without_excluded(
    original_union: Any, subset_union: Any, excluded: Iterable[Any]
) -> None:
    """
    Check that the ``subset_union`` ∪ ``excluded`` is ``original_union``.

    Make sure you put the assertion at the end of the module where no new classes are
    defined.

    See also for more details: https://hakibenita.com/python-mypy-exhaustive-checking
    """
    # region Map the identifiers of the inputs to their objects

    if inspect.isclass(original_union):
        original_union_map = {id(original_union): original_union}
    elif hasattr(original_union, "__args__"):
        original_union_map = {id(a_type): a_type for a_type in original_union.__args__}
    else:
        raise NotImplementedError(
            f"We do not know how to handle the original_union: {original_union}"
        )

    if inspect.isclass(subset_union):
        subset_union_map = {id(subset_union): subset_union}
    elif hasattr(subset_union, "__args__"):
        subset_union_map = {id(a_type): a_type for a_type in subset_union.__args__}
    else:
        raise NotImplementedError(
            f"We do not know how to handle the subset_union: {subset_union}"
        )

    excluded_map = {id(a_type): a_type for a_type in excluded}

    name_map = {**original_union_map, **subset_union_map, **excluded_map}

    # endregion

    # region Compute the sets of the identifiers

    original_union_set = set(original_union_map.keys())
    subset_union_set = set(subset_union_map.keys())
    excluded_set = set(excluded_map.keys())

    # endregion

    # region Check that it all fits

    intersection = subset_union_set.intersection(excluded_set)
    if len(intersection) > 0:
        names = sorted(name_map[type_id].__name__ for type_id in intersection)
        raise AssertionError(
            f"The following types were listed both "
            f"in the subset_union and the excluded: {names}"
        )

    diff = subset_union_set.difference(original_union_set)
    if len(diff) > 0:
        names = sorted(name_map[type_id].__name__ for type_id in diff)
        raise AssertionError(
            f"The following types were listed in the subset_union, "
            f"but not in the original_union: {names}"
        )

    diff = excluded_set.difference(original_union_set)
    if len(diff) > 0:
        names = sorted(name_map[type_id].__name__ for type_id in diff)
        raise AssertionError(
            f"The following types were listed in the excluded, "
            f"but not in the original_union: {names}"
        )

    reconstructed_set = subset_union_set.union(excluded_set)

    diff = original_union_set.difference(reconstructed_set)
    if len(diff) > 0:
        names = sorted(name_map[type_id].__name__ for type_id in diff)
        raise AssertionError(
            f"The following types were listed in the original_union, "
            f"but not in the subset_union ∪ excluded: {names}"
        )

    diff = reconstructed_set.difference(original_union_set)
    if len(diff) > 0:
        names = sorted(name_map[type_id].__name__ for type_id in diff)
        raise AssertionError(
            f"The following types were listed in the subset_union ∪ excluded, "
            f"but not in the original_union: {names}"
        )

    # endregion


@ensure(lambda text, result: text == "".join(result))
def wrap_text_into_lines(text: str, line_width: int = 60) -> List[str]:
    """
    Wrap the ``text`` into multiple lines.

    The tokens are split based on the whitespace. We make sure the articles are not
    left hanging between the lines. A line should observe the ``line_limit``,
    if possible.

    No new lines are added — the description should be given to the user in the original
    formatting. We merely split it in string literals for better code readability.
    """
    parts = text.split(" ")
    if len(parts) == 1:
        return [text]

    # NOTE (mristin, 2022-04-08):
    # We do not want to cut out "the", "a" and "an" on separate lines, so we split
    # the text once more in tokens where the articles are kept in the same token as
    # the word.
    tokens = []  # type: List[str]

    article = None  # type: Optional[str]
    for part in parts:
        if article is None:
            if part in ("a", "an", "the"):
                article = part
                continue
            else:
                tokens.append(part)
        else:
            if part in ("a", "an", "the"):
                # Append the previously observed ``article``;
                # the ``part`` becomes a new article.
                tokens.append(article)
                article = part
                continue

            tokens.append(f"{article} {part}")
            article = None

    if article is not None:
        tokens.append(article)

    # NOTE (mristin, 2022-04-08):
    # We add space to the tokens so that it is easier to re-flow them.
    tokens = [
        f"{token} " if i < len(tokens) - 1 else token for i, token in enumerate(tokens)
    ]
    assert "".join(tokens) == text

    segments = []  # type: List[str]

    accumulation_len = 0
    accumulation = []  # type: List[str]

    for token in tokens:
        if len(token) > line_width:
            segments.append("".join(accumulation))
            segments.append(token)
            accumulation_len = 0
            accumulation = []

        elif accumulation_len + len(token) > line_width:
            segments.append("".join(accumulation))
            accumulation_len = len(token)
            accumulation = [token]
        else:
            accumulation_len += len(token)
            accumulation.append(token)

    if accumulation_len > 0:
        segments.append("".join(accumulation))

    return segments


T = TypeVar("T")


def pairwise(iterable: Iterable[T]) -> Iterator[Tuple[T, T]]:
    """
    Iterate pair-wise over the iterator.

    >>> list(pairwise("ABCDE"))
    [('A', 'B'), ('B', 'C'), ('C', 'D'), ('D', 'E')]
    """
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def iterate_except_first(iterable: Iterable[T]) -> Iterator[T]:
    """
    Iterate over ``iterable``, but skip the first item.

    >>> list(iterate_except_first("ABCD"))
    ['B', 'C', 'D']
    """
    iterator = iter(iterable)
    next(iterator, None)
    yield from iterator



"""Generate names from ``Our_identifier_snake_case`` to other cases."""

from typing import List

from icontract import ensure, require

from aas_core_codegen.common import Identifier


def lower_snake_case(identifier: Identifier) -> Identifier:
    """Convert the identifier to a ``lower_snake_case``."""
    parts = identifier.split("_")

    assert len(parts) > 0, "Expected at least one part in the identifier"

    return Identifier("_".join(part.lower() for part in parts))


def upper_snake_case(identifier: Identifier) -> Identifier:
    """Convert the identifier to a ``UPPER_SNAKE_CASE``."""
    parts = identifier.split("_")

    assert len(parts) > 0, "Expected at least one part in the identifier"

    return Identifier("_".join(part.upper() for part in parts))


def lower_camel_case(identifier: Identifier) -> Identifier:
    """Convert the identifier to a ``camelCase``."""
    parts = identifier.split("_")

    assert len(parts) > 0, "Expected at least one part in the identifier"

    if len(parts) == 1:
        return Identifier(parts[0].lower())

    cased_parts = []  # type: List[str]

    iterator = iter(parts)
    first_part = next(iterator)
    cased_parts.append(first_part.lower())

    for part in iterator:
        cased_parts.append(part.capitalize())

    return Identifier("".join(cased_parts))


def capitalized_camel_case(identifier: Identifier) -> Identifier:
    """Convert the identifier to a ``CamelCase``."""
    parts = identifier.split("_")
    return Identifier("".join(part.capitalize() for part in parts))


def json_property(identifier: Identifier) -> Identifier:
    """
    Generate a JSON name of a property based on its meta-model ``identifier``.

    >>> json_property(Identifier("something"))
    'something'

    >>> json_property(Identifier("something_to_URL"))
    'somethingToUrl'

    >>> json_property(Identifier("global_asset_ID"))
    'globalAssetId'

    >>> json_property(Identifier("specific_asset_IDs"))
    'specificAssetIds'
    """
    return lower_camel_case(identifier)


# fmt: off
@require(
    lambda identifier: identifier[0].isupper(),
    "The class name must start with a capital letter"
)
@ensure(
    lambda result: "_" not in result
) # This post-condition avoids naming conflicts with prefixing in the JSON schema.
@ensure(
    lambda result: '"' not in result and "'" not in result and '\\' not in result
    # This post-condition simplifies a lot of generator code since we can simply in-line
    # the model type in strings.
)
# fmt: on
def json_model_type(identifier: Identifier) -> Identifier:
    """
    Generate the ``modelType`` of the class based on its meta-model ``identifier``.

    >>> json_model_type(Identifier("Something"))
    'Something'

    >>> json_model_type(Identifier("Data_type_IEC_61360"))
    'DataTypeIec61360'

    >>> json_model_type(Identifier("URL_to_something"))
    'UrlToSomething'
    """
    return capitalized_camel_case(identifier)


# fmt: off
@require(
    lambda identifier: identifier[0].upper() == identifier[0],
    "The class name must start with a capital letter"
)
# fmt: on
def xml_class_name(identifier: Identifier) -> Identifier:
    """
    Generate the XML tag name for the given class based on its ``identifier``.

    >>> xml_class_name(Identifier("Something"))
    'something'

    >>> xml_class_name(Identifier("URL_to_something"))
    'urlToSomething'
    """
    return lower_camel_case(identifier)


def xml_property(identifier: Identifier) -> Identifier:
    """
    Generate the XML name for the given property based on its ``identifier``.

    >>> xml_property(Identifier("something"))
    'something'

    >>> xml_property(Identifier("URL_to_something"))
    'urlToSomething'
    """
    return lower_camel_case(identifier)



"""Generate the Golang code for XML-ization based on the intermediate representation."""

import io
from typing import Tuple, Optional, List

from icontract import ensure, require

from aas_core_codegen import intermediate, naming, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    assert_never,
    indent_but_first_line,
)
from aas_core_codegen.golang import (
    common as golang_common,
    naming as golang_naming,
    pointering as golang_pointering,
)
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
)


# region De-serialization


def _generate_deserialization_error_and_its_methods() -> List[Stripped]:
    """Generate the code to represent the deserialization error."""
    return [
        Stripped(
            f"""\
// Represent an error during the de-serialization.
//
// Implements `error`.
type DeserializationError struct{{
{I}Path *aasreporting.Path
{I}Message string
}}"""
        ),
        Stripped(
            f"""\
func newDeserializationError(message string) *DeserializationError {{
{I}return &DeserializationError{{
{II}Path: &aasreporting.Path{{}},
{II}Message: message,
{I}}}
}}"""
        ),
        Stripped(
            f"""\
func (de *DeserializationError) Error() string {{
{I}return fmt.Sprintf(
{II}"%s: %s",
{II}de.PathString(),
{II}de.Message,
{I})
}}"""
        ),
        Stripped(
            f"""\
// Render the path as a string.
func (de *DeserializationError) PathString() string {{
{I}return aasreporting.ToRelativeXPath(de.Path)
}}"""
        ),
    ]


def _generate_is_whitespace() -> Stripped:
    return Stripped(
        f"""\
// Check if the string `s` consists only of whitespace.
//
// An empty string causes panic — please cover that case before.
func isWhitespace(s string) bool {{
{I}if len(s) == 0 {{
{II}panic("Unexpected empty string")
{I}}}
{I}for _, c := range s {{
{II}if !unicode.IsSpace(c) {{
{III}return false
{II}}}
{I}}}
{I}return true
}}"""
    )


def _generate_read_next() -> Stripped:
    return Stripped(
        f"""\
// Read the next token from the `decoder` given the `current` token.
//
// If `current` token is [eof], return [eof].
func readNext(decoder *xml.Decoder, current xml.Token) (next xml.Token, err error) {{
{I}if _, isEOF := current.(eof); isEOF {{
{II}next = current
{II}return
{I}}}

{I}var tokenErr error
{I}next, tokenErr = decoder.Token()
{I}if tokenErr != nil {{
{II}if tokenErr == io.EOF {{
{III}next = &eof{{}}
{III}return
{II}}}

{II}err = tokenErr
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_skip_empty_text_whitespace_and_comments() -> Stripped:
    return Stripped(
        f"""\
// Read all the possible whitespace and comments.
//
// Return the `next` token which is neither empty text, nor whitespace nor comment,
// or [eof], if we reached the end-of-file.
//
// If we already reached the end-of-file, simply return [eof].
func skipEmptyTextWhitespaceAndComments(
{I}decoder *xml.Decoder,
{I}current xml.Token,
) (next xml.Token, err error) {{
{I}stop := false
{I}for !stop {{
{II}if _, isEOF := current.(eof); isEOF {{
{III}break
{II}}}

{II}switch et := current.(type) {{
{II}case xml.CharData:
{III}text := string(et)
{III}if len(text) != 0 && !isWhitespace(text) {{
{IIII}stop = true
{III}}} else {{
{IIII}// We should proceed to the next token.
{III}}}
{II}case xml.Comment:
{III}// We should proceed to the next token.
{II}default:
{III}stop = true
{II}}}

{II}if !stop {{
{III}current, err = readNext(decoder, current)
{III}if err != nil {{
{IIII}return
{III}}}
{II}}}
{I}}}

{I}next = current
{I}return
}}"""
    )


def _generate_read_text() -> Stripped:
    return Stripped(
        f"""\
// Consume the text tokens (char data).
//
// Any comment tokens are skipped.
//
// The resulting `next` token points to the first token which is neither text
// nor comment.
//
// If we reached the end-of-file, `next` is an [eof] sentinel token.
func readText(
{I}decoder *xml.Decoder,
{I}current xml.Token,
) (text string, next xml.Token, err error) {{
{I}b := &strings.Builder{{}}

{I}stop := false
{I}for {{
{II}if _, isEOF := current.(eof); isEOF {{
{III}err = newDeserializationError(
{IIII}"Expected to read text, but reached the end-of-file",
{III})
{III}return
{II}}}

{II}switch et := current.(type) {{
{II}case xml.CharData:
{III}b.WriteString(string(et))
{III}// Proceed to the next token.
{II}case xml.Comment:
{III}// Proceed to the next token.
{II}default:
{III}stop = true
{II}}}

{II}if !stop {{
{III}current, err = readNext(decoder, current)
{III}if err != nil {{
{IIII}return
{III}}}
{II}}} else {{
{III}break
{II}}}
{I}}}

{I}next = current
{I}text = b.String()
{I}return
}}"""
    )


def _generate_read_text_as_boolean() -> Stripped:
    return Stripped(
        f"""\
// Consume the text tokens (char data) as a representation of a `xs:boolean`.
//
// Any comment tokens are skipped.
//
// The resulting `next` token points to the first token which is neither text
// nor comment.
//
// If we reached the end-of-file, `next` is an [eof] sentinel token.
func readTextAsBoolean(
{I}decoder *xml.Decoder,
{I}current xml.Token,
) (value bool, next xml.Token, err error) {{
{I}var text string
{I}text, next, err = readText(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}switch text {{
{I}case "1":
{II}value = true
{I}case "true":
{II}value = true
{I}case "0":
{II}value = false
{I}case "false":
{II}value = false
{I}default:
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected a value as xs:boolean, but got: %s",
{IIII}text,
{III}),
{II})
{I}}}
{I}if err != nil {{
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_read_text_as_long() -> Stripped:
    return Stripped(
        f"""\
// Consume the text tokens (char data) as a representation of a `xs:long`.
//
// Any comment tokens are skipped.
//
// The resulting `next` token points to the first token which is neither text
// nor comment.
//
// If we reached the end-of-file, `next` is an [eof] sentinel token.
func readTextAsLong(
{I}decoder *xml.Decoder,
{I}current xml.Token,
) (value int64, next xml.Token, err error) {{
{I}var text string
{I}text, next, err = readText(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}value, err = strconv.ParseInt(text, 10, 64)

{I}return
}}"""
    )


def _generate_read_text_as_double() -> Stripped:
    return Stripped(
        f"""\
// Consume the text tokens (char data) as a representation of a `xs:double`.
//
// Any comment tokens are skipped.
//
// The resulting `next` token points to the first token which is neither text
// nor comment.
//
// If we reached the end-of-file, `next` is an [eof] sentinel token.
func readTextAsDouble(
{I}decoder *xml.Decoder,
{I}current xml.Token,
) (value float64, next xml.Token, err error) {{
{I}var text string
{I}text, next, err = readText(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}// We need to check explicitly for the regular expression since
{I}// strconv.ParseFloat is too permissive. For example, it accepts "nan"
{I}// although only "NaN" is valid.
{I}// See: https://www.w3.org/TR/xmlschema-2/#double
{I}if !aasverification.MatchesXsDouble(text) {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected a value as xs:double, but got: %s",
{IIII}text,
{III}),
{II})
{II}return
{I}}}

{I}var parseErr error
{I}value, parseErr = strconv.ParseFloat(text, 64)
{I}if parseErr != nil {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected a value as xs:double, but it could not be parsed: %s: %s",
{IIII}parseErr.Error(), text,
{III}),
{II})
{II}return
{I}}}

{I}// NOTE (2023-06-14):
{I}// We explicitly do not check for loss of precision, as the majority of people will
{I}// use string representation of the floating point numbers ignoring the precision
{I}// issues. For example, the closest double-precision number to the number `359.9` is
{I}// `359.8999999999999772626324556767940521240234375`, but most people will simply
{I}// give `359.9` as the value.

{I}return
}}"""
    )


def _generate_read_text_as_base64_encoded_bytes() -> Stripped:
    return Stripped(
        f"""\
// Consume the text tokens (char data) as a base64-encoded bytes.
//
// Any comment tokens are skipped.
//
// The resulting `next` token points to the first token which is neither text
// nor comment.
//
// If we reached the end-of-file, `next` is an [eof] sentinel token.
func readTextAsBase64EncodedBytes(
{I}decoder *xml.Decoder,
{I}current xml.Token,
) (value []byte, next xml.Token, err error) {{
{I}var text string
{I}text, next, err = readText(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}var decodingErr error
{I}value, decodingErr = b64.StdEncoding.DecodeString(text)
{I}if decodingErr != nil {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Text could not be decoded as base64: %s",
{IIII}decodingErr.Error(),
{III}),
{II})
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_check_start_element() -> Stripped:
    return Stripped(
        f"""\
// Check that the `current` token is a valid start element, *i.e.*, lives in [Namespace]
// and contains no attributes.
func checkStartElement(
{I}current xml.StartElement,
) (err error) {{
{I}const xmlnsLen = len("xmlns")

{I}unexpectedAttr := 0
{I}for _, attr := range current.Attr {{
{II}if (attr.Name.Space == "" && attr.Name.Local == "xmlns") ||
{III}attr.Name.Space == "xmlns" {{
{III}continue
{II}}}

{II}unexpectedAttr++
{I}}}
{I}if unexpectedAttr != 0 {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected no attributes except 'xmlns' in the start element, "+
{IIIII}"but got %d in the start element %s",
{IIII}unexpectedAttr, current.Name.Local,
{III}),
{II})
{II}return
{I}}}

{I}if current.Name.Space != Namespace {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected only start elements in the namespace %s, "+
{IIIII}"but got a start element %s in the namespace %s",
{IIII}Namespace, current.Name.Local, current.Name.Space,
{III}),
{II})
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_extract_local_name_from_start_element() -> Stripped:
    return Stripped(
        f"""\
// Expect a valid start element (as defined in [checkStartElement]) and extract its
// `local` name.
//
// This function is meant to be called whenever you know the runtime type of a token.
// If you do not know the runtime type, call [parseAsStartElementAndExtractLocalName]
// so that you can succinctly check the runtime type as well.
func extractLocalNameFromStartElement(
{I}current xml.StartElement,
) (local string, err error) {{
{I}err = checkStartElement(current)
{I}if err != nil {{
{II}return
{I}}}

{I}local = current.Name.Local
{I}return
}}"""
    )


def _generate_parse_as_start_element_and_extract_local_name() -> Stripped:
    return Stripped(
        f"""\
// Expect a valid start element (as defined in [checkStartElement]) and extract its
// local name.
//
// Valid means that we check that the start element lives in [Namespace] and contains
// no attributes.
//
// If you know the runtime type of `current` token, call
// [parseLocalNameFromStartElement] instead to save a cast.
func parseAsStartElementAndExtractLocalName(
{I}current xml.Token,
) (local string, err error) {{
{I}if _, isEOF := current.(eof); isEOF {{
{II}err = newDeserializationError(
{III}"Expected a start element, but reached the end-of-file",
{II})
{II}return
{I}}}

{I}et, ok := current.(xml.StartElement)
{I}if !ok {{
{II}switch v := current.(type) {{
{II}case xml.EndElement:
{III}err = newDeserializationError(
{IIII}fmt.Sprintf(
{IIIII}"Expected a start element, but got an end element %s in namespace %s",
{IIIII}v.Name.Local, v.Name.Space,
{IIII}),
{III})
{II}case xml.CharData:
{III}err = newDeserializationError(
{IIII}fmt.Sprintf(
{IIIII}"Expected a start element, but got text %s",
{IIIII}string(v),
{IIII}),
{III})
{II}default:
{III}err = newDeserializationError(
{IIII}fmt.Sprintf(
{IIIII}"Expected a start element, but got %T: %v",
{IIIII}current, current,
{IIII}),
{III})
{II}}}
{II}return
{I}}}

{I}local, err = extractLocalNameFromStartElement(et)
{I}return
}}"""
    )


def _generate_check_end_element() -> Stripped:
    return Stripped(
        f"""\
// Check that the `current` token is an end element, living in [Namespace], and
// having the `local` name.
func checkEndElement(current xml.Token, local string) (err error) {{
{I}if _, isEOF := current.(eof); isEOF {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected an end element %s, but reached the end-of-file",
{IIII}local,
{III}),
{II})
{II}return
{I}}}

{I}et, ok := current.(xml.EndElement)
{I}if !ok {{
{II}switch v := current.(type) {{
{II}case xml.StartElement:
{III}err = newDeserializationError(
{IIII}fmt.Sprintf(
{IIIII}"Expected an end element %s, but got a start element %s in namespace %s",
{IIIII}local, v.Name.Local, v.Name.Space,
{IIII}),
{III})
{II}case xml.CharData:
{III}err = newDeserializationError(
{IIII}fmt.Sprintf(
{IIIII}"Expected an end element %s, but got text %s",
{IIIII}local, string(v),
{IIII}),
{III})
{II}default:
{III}err = newDeserializationError(
{IIII}fmt.Sprintf(
{IIIII}"Expected an end element %s, but got %T: %v",
{IIIII}local, current, current,
{IIII}),
{III})
{II}}}
{II}return
{I}}}

{I}if et.Name.Space != Namespace {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected an end element %s in the namespace %s, "+
{IIIII}"but got an end element in the namespace %s",
{IIII}local, Namespace, et.Name.Space,
{III}),
{II})
{II}return
{I}}}

{I}if et.Name.Local != local {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected an end element %s, but got an end element %s",
{IIII}local, et.Name.Local,
{III}),
{II})
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_read_list() -> Stripped:
    return Stripped(
        f"""\
// Read a list of AAS instances as a sequence of XML elements.
//
// Every start element is considered to mark the start of an instance serialization. We
// stop the reading as soon as we encounter a non-start element.
//
// That last non-start element is returned as `next` element.
func readList[T aastypes.IClass](
{I}decoder *xml.Decoder,
{I}current xml.Token,
{I}readTWithLookahead func(
{II}aDecoder *xml.Decoder,
{II}aCurrent xml.Token,
{I}) (anInstance T, anErr error),
) (instances []T, next xml.Token, err error) {{
{I}i := 0
{I}for {{
{II}current, err = skipEmptyTextWhitespaceAndComments(decoder, current)
{II}if err != nil {{
{III}return
{II}}}

{II}if _, ok := current.(xml.StartElement); !ok {{
{III}break
{II}}}

{II}var instance T
{II}var instanceErr error
{II}instance, instanceErr = readTWithLookahead(decoder, current)
{II}if instanceErr != nil {{
{III}if deseriaErr, ok := instanceErr.(*DeserializationError); ok {{
{IIII}deseriaErr.Path.PrependIndex(
{IIIII}&aasreporting.IndexSegment{{Index: i}},
{IIII})
{III}}}
{III}err = instanceErr
{III}return
{II}}}

{II}instances = append(instances, instance)

{II}i++

{II}current, err = readNext(decoder, nil)
{II}if err != nil {{
{III}return
{II}}}
{I}}}

{I}next = current
{I}return
}}"""
    )


def _generate_read_text_as_enumeration(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    enum_name = golang_naming.enum_name(enumeration.name)
    from_string_name = golang_naming.function_name(
        Identifier(f"{enumeration.name}_from_string")
    )

    function_name = golang_naming.private_function_name(
        Identifier(f"read_text_as_{enumeration.name}")
    )

    return Stripped(
        f"""\
// Consume the text tokens (char data) as a string-encoded literal of
// [aastypes.{enum_name}].
//
// Any comment tokens are skipped.
//
// The resulting `next` token points to the first token which is neither text
// nor comment.
//
// If we reached the end-of-file, `next` is an [eof] sentinel token.
func {function_name}(
{I}decoder *xml.Decoder,
{I}current xml.Token,
) (value aastypes.{enum_name},
{I}next xml.Token,
{I}err error,
) {{
{I}var text string
{I}text, next, err = readText(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}var ok bool
{I}value, ok = aasstringification.{from_string_name}(text)
{I}if !ok {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Unexpected literal of {enum_name}: %v",
{IIII}text,
{III}),
{II})
{II}return
{I}}}

{I}return
}}"""
    )


_READ_FUNCTION_BY_PRIMITIVE_TYPE = {
    intermediate.PrimitiveType.BOOL: "readTextAsBoolean",
    intermediate.PrimitiveType.INT: "readTextAsLong",
    intermediate.PrimitiveType.FLOAT: "readTextAsDouble",
    intermediate.PrimitiveType.STR: "readText",
    intermediate.PrimitiveType.BYTEARRAY: "readTextAsBase64EncodedBytes",
}
assert all(
    literal in _READ_FUNCTION_BY_PRIMITIVE_TYPE
    for literal in intermediate.PrimitiveType
)


def _generate_snippet_to_switch_on_property_deserialization(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """
    Generate the switch block to dispatch how to read a property.

    The start element is expected to have been read. The variable ``local`` denotes
    the local name of the start element.

    The decoder points to the first token of the property content.

    The variables ``the*`` and ``found*`` will be set as well as ``valueErr``.
    """
    case_blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        type_anno = intermediate.beneath_optional(prop.type_annotation)

        prop_var = golang_naming.variable_name(Identifier(f"the_{prop.name}"))

        xml_prop_literal = golang_common.string_literal(naming.xml_property(prop.name))

        case_body_blocks = []  # type: List[Stripped]

        if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation) or (
            isinstance(type_anno, intermediate.OurTypeAnnotation)
            and isinstance(
                type_anno.our_type,
                (intermediate.ConstrainedPrimitive, intermediate.Enumeration),
            )
        ):
            primitive_type = intermediate.try_primitive_type(type_anno)

            if primitive_type is not None:
                read_function = _READ_FUNCTION_BY_PRIMITIVE_TYPE[primitive_type]
            else:
                assert isinstance(
                    type_anno, intermediate.OurTypeAnnotation
                ) and isinstance(type_anno.our_type, intermediate.Enumeration)

                read_function = golang_naming.private_function_name(
                    Identifier(f"read_text_as_{type_anno.our_type.name}")
                )

            pointer = golang_pointering.is_pointer_type(prop.type_annotation)

            if pointer:
                # NOTE (mristin, 2023-06-17):
                # We explicitly pass in ``type_anno`` to the type as the read function
                # will return a non-optional.
                value_type = golang_common.generate_type(
                    type_annotation=type_anno, types_package=Identifier("aastypes")
                )

                case_body_blocks.append(
                    Stripped(
                        f"""\
var value {value_type}
value, current, valueErr = {read_function}(
{I}decoder,
{I}current,
)
{prop_var} = &value"""
                    )
                )
            else:
                case_body_blocks.append(
                    Stripped(
                        f"""\
{prop_var}, current, valueErr = {read_function}(
{I}decoder,
{I}current,
)"""
                    )
                )

        elif isinstance(type_anno, intermediate.OurTypeAnnotation):
            our_type = type_anno.our_type

            if isinstance(our_type, intermediate.Enumeration):
                raise AssertionError("Must have been handled before")

            elif isinstance(our_type, intermediate.ConstrainedPrimitive):
                raise AssertionError("Must have been handled before")

            elif isinstance(
                our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
            ):
                if (
                    isinstance(our_type, intermediate.ConcreteClass)
                    and len(our_type.concrete_descendants) == 0
                ):
                    read_function = golang_naming.private_function_name(
                        Identifier(f"read_{type_anno.our_type.name}_as_sequence")
                    )

                    case_body_blocks.append(
                        Stripped(
                            f"""\
{prop_var}, current, valueErr =  {read_function}(
{I}decoder,
{I}current,
)"""
                        )
                    )

                else:
                    read_with_lookahead_function = golang_naming.private_function_name(
                        Identifier(f"read_{type_anno.our_type.name}_with_lookahead")
                    )

                    case_body_blocks.append(
                        Stripped(
                            f"""\
{prop_var}, valueErr =  {read_with_lookahead_function}(
{I}decoder,
{I}current,
)
// {read_with_lookahead_function} stops at the end element,
// so we look ahead to the next element, just after the end element.
if valueErr == nil {{
{I}current, valueErr = readNext(decoder, current)
}}"""
                        )
                    )

            else:
                assert_never(our_type)

        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert isinstance(
                type_anno.items, intermediate.OurTypeAnnotation
            ) and isinstance(
                type_anno.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ), (
                f"NOTE (mristin, 2023-03-29): We expect only lists of classes "
                f"at the moment, but you specified {type_anno}. "
                f"Please contact the developers if you need this feature."
            )

            read_item_function = golang_naming.private_function_name(
                Identifier(f"read_{type_anno.items.our_type.name}_with_lookahead")
            )

            case_body_blocks.append(
                Stripped(
                    f"""\
{prop_var}, current, valueErr = readList(
{I}decoder,
{I}current,
{I}{read_item_function},
)"""
                )
            )
        else:
            assert_never(type_anno)

        assert len(case_body_blocks) > 0

        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            found_var = golang_naming.variable_name(Identifier(f"found_{prop.name}"))
            case_body_blocks.append(Stripped(f"{found_var} = true"))

        case_body = "\n".join(case_body_blocks)
        case_blocks.append(
            Stripped(
                f"""\
case {xml_prop_literal}:
{I}{indent_but_first_line(case_body, I)}"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}valueErr = newDeserializationError(
{II}fmt.Sprintf(
{III}"Unexpected property",
{II}),
{I})"""
        )
    )

    case_blocks_joined = "\n\n".join(case_blocks)

    return Stripped(
        f"""\
var valueErr error
switch local {{
{case_blocks_joined}
}}"""
    )


def _generate_read_as_sequence(cls: intermediate.ConcreteClass) -> Stripped:
    interface_name = golang_naming.interface_name(cls.name)

    # region Initialize

    initialization_blocks = []  # type: List[Stripped]

    prop_var_initializations = []  # type: List[Stripped]
    for prop in cls.properties:
        prop_var = golang_naming.variable_name(Identifier(f"the_{prop.name}"))

        prop_var_type = golang_common.generate_type(
            type_annotation=prop.type_annotation, types_package=Identifier("aastypes")
        )

        prop_var_initializations.append(Stripped(f"var {prop_var} {prop_var_type}"))

    if len(prop_var_initializations) > 0:
        initialization_blocks.append(Stripped("\n".join(prop_var_initializations)))

    found_var_initializations = []  # type: List[Stripped]
    for prop in cls.properties:
        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        found_var = golang_naming.variable_name(Identifier(f"found_{prop.name}"))

        found_var_initializations.append(Stripped(f"{found_var} := false"))

    if len(found_var_initializations) > 0:
        initialization_blocks.append(Stripped("\n".join(found_var_initializations)))

    if len(initialization_blocks) == 0:
        initialization_blocks.append(
            Stripped(
                f"""\
// No initialization as there are no properties
// in {interface_name}."""
            )
        )

    initialization = "\n\n".join(initialization_blocks)

    # endregion

    switch_snippet = _generate_snippet_to_switch_on_property_deserialization(cls=cls)

    # region Construct

    construct_blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        found_var = golang_naming.variable_name(Identifier(f"found_{prop.name}"))

        message_literal = golang_common.string_literal(
            f"The required property {naming.json_property(prop.name)!r} is missing"
        )

        construct_blocks.append(
            Stripped(
                f"""\
if !{found_var} {{
{I}err = newDeserializationError(
{II}{message_literal},
{I})
{I}return
}}"""
            )
        )

    constructing_statements = []  # type: List[Stripped]

    constructor_arguments = [
        golang_naming.variable_name(Identifier(f"the_{arg.name}"))
        for arg in cls.constructor.arguments
        if not isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation)
    ]  # type: List[Stripped]

    new_function = golang_naming.function_name(Identifier(f"new_{cls.name}"))

    if len(constructor_arguments) > 0:
        constructor_arguments_joined = "\n".join(
            f"{arg}," for arg in constructor_arguments
        )

        constructing_statements.append(
            Stripped(
                f"""\
instance = aastypes.{new_function}(
{I}{indent_but_first_line(constructor_arguments_joined, I)}
)"""
            )
        )
    else:
        constructing_statements.append(
            Stripped(f"instance = aastypes.{new_function}()")
        )

    for arg in cls.constructor.arguments:
        if not isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        setter_name = golang_naming.setter_name(arg.name)
        prop_var = golang_naming.variable_name(Identifier(f"the_{arg.name}"))

        constructing_statements.append(
            Stripped(
                f"""\
instance.{setter_name}(
{I}{prop_var},
)"""
            )
        )

    construct_blocks.append(Stripped("\n".join(constructing_statements)))

    construct = "\n\n".join(construct_blocks)

    # endregion

    function_name = golang_naming.private_function_name(
        Identifier(f"read_{cls.name}_as_sequence")
    )

    return Stripped(
        f"""\
// De-serialize the instance of [aastypes.{interface_name}]
// as a sequence of XML elements, each representing a property
// of [aastypes.{interface_name}].
//
// The reading stops as soon as we encounter a non-start element, and we return
// that token as the `next` token.
func {function_name}(
{I}decoder *xml.Decoder,
{I}current xml.Token,
) (instance aastypes.{interface_name},
{I}next xml.Token,
{I}err error,
) {{
{I}{indent_but_first_line(initialization, I)}

{I}for {{
{II}current, err = skipEmptyTextWhitespaceAndComments(decoder, current)
{II}if err != nil {{
{III}return
{II}}}

{II}if _, isEOF := current.(eof); isEOF {{
{III}break
{II}}}

{II}startElement, ok := current.(xml.StartElement)
{II}if !ok {{
{III}if charData, isCharData := current.(xml.CharData); isCharData {{
{IIII}err = newDeserializationError(
{IIIII}fmt.Sprintf(
{IIIIII}"Expected a sequence of XML elements representing properties "+
{IIIIII}"of {interface_name}, but got text: %s",
{IIIIII}string(charData),
{IIIII}),
{IIII})
{IIII}return
{III}}}

{III}break
{II}}}

{II}var local string
{II}local, err = extractLocalNameFromStartElement(startElement)
{II}if err != nil {{
{III}return
{II}}}

{II}// Move the current to the content of the XML element
{II}current, err = readNext(decoder, nil)
{II}if err != nil {{
{III}return
{II}}}

{II}{indent_but_first_line(switch_snippet, II)}

{II}if valueErr != nil {{
{III}if deseriaErr, ok := valueErr.(*DeserializationError); ok {{
{IIII}deseriaErr.Path.PrependName(
{IIIII}&aasreporting.NameSegment{{Name: local}},
{IIII})
{III}}}
{III}err = valueErr
{II}}}

{II}if err != nil {{
{III}return
{II}}}

{II}current, err = skipEmptyTextWhitespaceAndComments(decoder, current)
{II}if err != nil {{
{III}return
{II}}}

{II}err = checkEndElement(current, local)
{II}if err != nil {{
{III}return
{II}}}

{II}current, err = readNext(decoder, current)
{II}if err != nil {{
{III}return
{II}}}
{I}}}

{I}current, err = skipEmptyTextWhitespaceAndComments(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}next = current

{I}{indent_but_first_line(construct, I)}
{I}return
}}"""
    )


def _generate_read_with_lookahead_without_dispatch(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    interface_name = golang_naming.interface_name(cls.name)
    function_name = golang_naming.private_function_name(
        Identifier(f"read_{cls.name}_with_lookahead")
    )

    xml_class_name_literal = golang_common.string_literal(
        naming.xml_class_name(cls.name)
    )

    read_as_sequence_name = golang_naming.private_function_name(
        Identifier(f"read_{cls.name}_as_sequence")
    )

    return Stripped(
        f"""\
// De-serialize an instance of [aastypes.{interface_name}]
// as an XML element where the start element is expected to have been already
// read as `current` token.
//
// The de-serialization stops by consuming the final end element. The next call to
// the `decoder.Token()` will return the element just after the end element.
func {function_name}(
{I}decoder *xml.Decoder,
{I}current xml.Token,
) (instance aastypes.{interface_name},
{I}err error,
) {{
{I}current, err = skipEmptyTextWhitespaceAndComments(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}var local string
{I}local, err = parseAsStartElementAndExtractLocalName(
{II}current,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}expectedLocal := {xml_class_name_literal}
{I}if local != expectedLocal {{
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Expected a start element with local name %s, "+
{IIIII}"but got a start element with local name %s",
{IIII}expectedLocal, local,
{III}),
{II})
{II}return
{I}}}

{I}current, err = readNext(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}instance, current, err = {read_as_sequence_name}(
{II}decoder,
{II}current,
)
{I}if err != nil {{
{II}return
{I}}}

{I}err = checkEndElement(current, local)
{I}return
}}"""
    )


@require(
    lambda cls: len(cls.concrete_descendants) > 0,
    "The class must have one or more concrete descendants; "
    "otherwise the dispatch makes no sense",
)
def _generate_read_with_lookahead_with_dispatch(
    cls: intermediate.ClassUnion,
) -> Stripped:
    interface_name = golang_naming.interface_name(cls.name)
    function_name = golang_naming.private_function_name(
        Identifier(f"read_{cls.name}_with_lookahead")
    )

    case_blocks = []  # type: List[Stripped]

    for descendant_cls in cls.concrete_descendants:
        xml_class_name_literal = golang_common.string_literal(
            naming.xml_class_name(descendant_cls.name)
        )
        read_as_sequence = golang_naming.private_function_name(
            Identifier(f"read_{descendant_cls.name}_as_sequence")
        )
        case_blocks.append(
            Stripped(
                f"""\
case {xml_class_name_literal}:
{I}instance, current, err = {read_as_sequence}(
{II}decoder, current,
{I})"""
            )
        )

    if isinstance(cls, intermediate.ConcreteClass):
        xml_class_name_literal = golang_common.string_literal(
            naming.xml_class_name(cls.name)
        )
        read_as_sequence = golang_naming.private_function_name(
            Identifier(f"read_{cls.name}_as_sequence")
        )
        case_blocks.append(
            Stripped(
                f"""\
case {xml_class_name_literal}:
{I}instance, current, err = {read_as_sequence}(
{II}decoder, current,
{I})"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}err = newDeserializationError(
{II}fmt.Sprintf(
{III}"Unexpected start element %s as discriminator "+
{IIII}"for {interface_name}",
{III}local,
{II}),
{I})"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    switch_stmt = Stripped(
        f"""\
switch local {{
{case_blocks_joined}
}}"""
    )

    return Stripped(
        f"""\
// De-serialize an instance of [aastypes.{interface_name}]
// as an XML element where the start element is expected to have been already read
// as `current` token.
//
// The de-serialization stops by consuming the final end element. The next call to
// the `decoder.Token()` will return the element just after the end element.
func {function_name}(
{I}decoder *xml.Decoder,
{I}current xml.Token,
) (instance aastypes.{interface_name},
{I}err error,
) {{
{I}current, err = skipEmptyTextWhitespaceAndComments(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}var local string
{I}local, err = parseAsStartElementAndExtractLocalName(
{II}current,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}// Move the current to the properties of the instance
{I}current, err = readNext(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}{indent_but_first_line(switch_stmt, I)}
{I}if err != nil {{
{II}return
{I}}}

{I}err = checkEndElement(current, local)
{I}return
}}"""
    )


def _generate_unmarshal(symbol_table: intermediate.SymbolTable) -> Stripped:
    case_blocks = []  # type: List[Stripped]
    for cls in symbol_table.concrete_classes:
        read_as_sequence = golang_naming.private_function_name(
            Identifier(f"read_{cls.name}_as_sequence")
        )

        xml_class_name_literal = golang_common.string_literal(
            naming.xml_class_name(cls.name)
        )

        case_blocks.append(
            Stripped(
                f"""\
case {xml_class_name_literal}:
{I}instance, current, err = {read_as_sequence}(
{II}decoder, current,
{I})"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{II}err = newDeserializationError(
{III}fmt.Sprintf(
{IIII}"Unexpected XML element name %s as class discriminator",
{IIII}local,
{III}),
{II})"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    return Stripped(
        f"""\
// Unmarshal an instance of [aastypes.IClass] serialized as an XML element.
//
// The XML element must live in the [Namespace] space.
func Unmarshal(
{I}decoder *xml.Decoder,
) (instance aastypes.IClass, err error) {{
{I}var current xml.Token
{I}current, err = readNext(decoder, nil)
{I}if err != nil {{
{II}return
{I}}}

{I}current, err = skipEmptyTextWhitespaceAndComments(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}var local string
{I}local, err = parseAsStartElementAndExtractLocalName(
{II}current,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}// Move the current to the properties of the instance
{I}current, err = readNext(decoder, current)
{I}if err != nil {{
{II}return
{I}}}

{I}switch local {{
{II}{indent_but_first_line(case_blocks_joined, II)}
{I}}}
{I}if err != nil {{
{II}return
{I}}}

{I}err = checkEndElement(current, local)
{I}return
}}"""
    )


# endregion

# region Serialization


def _generate_serialization_error() -> List[Stripped]:
    return [
        Stripped(
            f"""\
// Represent an error during the serialization.
//
// Implements `error`.
type SerializationError struct {{
{I}Path    *aasreporting.Path
{I}Message string
}}"""
        ),
        Stripped(
            f"""\
func newSerializationError(message string) *SerializationError {{
{I}return &SerializationError{{
{II}Path:    &aasreporting.Path{{}},
{II}Message: message,
{I}}}
}}"""
        ),
        Stripped(
            f"""\
func (se *SerializationError) Error() string {{
{I}return fmt.Sprintf(
{II}"%s: %s",
{II}se.PathString(),
{II}se.Message,
{I})
}}"""
        ),
        Stripped(
            f"""\
// Render the path as a string.
func (se *SerializationError) PathString() string {{
{I}return aasreporting.ToGolangPath(se.Path)
}}"""
        ),
    ]


def _generate_write_start_element() -> Stripped:
    return Stripped(
        f"""\
// Write the start element with the given `local` name to the encoder.
//
// Do not flush.
//
// If the `withNamespace` is set, set the [xml.Name.Space] property in the element
// accordingly.
func writeStartElement(
{I}encoder *xml.Encoder,
{I}local string,
{I}withNamespace bool,
) (err error) {{
{I}startElement := xml.StartElement{{Name: xml.Name{{Local: local}}}}
{I}if withNamespace {{
{II}startElement.Name.Space = Namespace
{I}}}

{I}err = encoder.EncodeToken(startElement)
{I}return
}}"""
    )


def _generate_write_end_element() -> Stripped:
    return Stripped(
        f"""\
// Write the end element with the given `local` name to the encoder.
//
// Do not flush.
//
// If the `withNamespace` is set, set the [xml.Name.Space] property in the element
// accordingly.
func writeEndElement(
{I}encoder *xml.Encoder,
{I}local string,
{I}withNamespace bool,
) (err error) {{
{I}endElement := xml.EndElement{{Name: xml.Name{{Local: local}}}}
{I}if withNamespace {{
{II}endElement.Name.Space = Namespace
{I}}}

{I}err = encoder.EncodeToken(endElement)
{I}return
}}"""
    )


def _generate_write_text() -> Stripped:
    return Stripped(
        f"""\
// Write the `text` to the encoder.
//
// Do not flush.
//
// If `text` is empty, do nothing.
func writeText(
{I}encoder *xml.Encoder,
{I}text string,
) (err error) {{
{I}if len(text) > 0 {{
{II}err = encoder.EncodeToken(
{III}xml.CharData([]byte(text)),
{II})
{I}}}
{I}return
}}"""
    )


def _generate_write_boolean_property() -> Stripped:
    return Stripped(
        f"""\
// Write the `value` of a property as `xs:boolean` enclosed in an XML element.
//
// Do not flush.
//
// The XML namespace is expected to have been defined outside of the resulting XML
// element.
func writeBooleanProperty(
{I}encoder *xml.Encoder,
{I}local string,
{I}value bool,
) (err error) {{
{I}err = writeStartElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}text := "true"
{I}if !value {{
{II}text = "false"
{I}}}
{I}err = writeText(encoder, text)
{I}if err != nil {{
{II}return
{I}}}

{I}err = writeEndElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_write_long_property() -> Stripped:
    return Stripped(
        f"""\
// Write the `value` of a property as `xs:long` enclosed in an XML element.
//
// Do not flush.
//
// The XML namespace is expected to have been defined outside of the resulting XML
// element.
func writeLongProperty(
{I}encoder *xml.Encoder,
{I}local string,
{I}value int64,
{I}withNamespace bool,
) (err error) {{
{I}err = writeStartElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}text := strconv.FormatInt(value, 10)
{I}err = writeText(encoder, text)
{I}if err != nil {{
{II}return
{I}}}

{I}err = writeEndElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_write_double_property() -> Stripped:
    return Stripped(
        f"""\
// Write the `value` of a property as `xs:double` enclosed in an XML element.
//
// Do not flush.
//
// The XML namespace is expected to have been defined outside of the resulting XML
// element.
func writeDoubleProperty(
{I}encoder *xml.Encoder,
{I}local string,
{I}value float64,
) (err error) {{
{I}err = writeStartElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}var text string

{I}// See: https://www.w3.org/TR/xmlschema-2/#double
{I}// for the exact literals.
{I}if math.IsInf(value, 0) {{
{II}if value < 0 {{
{III}text = "-INF"
{II}}} else {{
{III}text = "INF"
{II}}}
{I}}} else if math.IsNaN(value) {{
{II}text = "NaN"
{I}}} else {{
{II}text = strconv.FormatFloat(value, 'g', -1, 64)
{I}}}

{I}err = writeText(encoder, text)
{I}if err != nil {{
{II}return
{I}}}

{I}err = writeEndElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_write_string_property() -> Stripped:
    return Stripped(
        f"""\
// Write the `value` of a property as `xs:string` enclosed in an XML element.
//
// Do not flush.
//
// The XML namespace is expected to have been defined outside of the resulting XML
// element.
func writeStringProperty(
{I}encoder *xml.Encoder,
{I}local string,
{I}value string,
) (err error) {{
{I}err = writeStartElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}err = writeText(encoder, value)
{I}if err != nil {{
{II}return
{I}}}

{I}err = writeEndElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_write_bytes_property() -> Stripped:
    return Stripped(
        f"""\
// Write the `value` of a property as base64-encoded bytes.
//
// Do not flush.
//
// The XML namespace is expected to have been defined outside of the resulting XML
// element.
func writeBytesProperty(
{I}encoder *xml.Encoder,
{I}local string,
{I}value []byte,
) (err error) {{
{I}err = writeStartElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}text := b64.StdEncoding.EncodeToString(
{II}value,
{I})

{I}err = writeText(encoder, text)
{I}if err != nil {{
{II}return
{I}}}

{I}err = writeEndElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_write_list_as_sequence() -> Stripped:
    return Stripped(
        f"""\
// Serialize the list of instances as a sequence of XML elements enclosed in a parent
// XML element with the `local` name.
func writeListProperty[T aastypes.IClass](
{I}encoder *xml.Encoder,
{I}local string,
{I}list []T,
) (err error) {{
{I}err = writeStartElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}for i, item := range list {{
{II}err = Marshal(
{III}encoder,
{III}item,
{III}false,
{II})
{II}if err != nil {{
{III}if seriaErr, ok := err.(*SerializationError); ok {{
{IIII}seriaErr.Path.PrependIndex(
{IIIII}&aasreporting.IndexSegment{{
{IIIIII}Index: i,
{IIIII}}},
{IIII})
{III}}}
{III}return
{II}}}
{I}}}

{I}err = writeEndElement(
{II}encoder,
{II}local,
{II}false,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}return
}}"""
    )


def _generate_write_enumeration_property(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    enum_name = golang_naming.enum_name(enumeration.name)
    function_name = golang_naming.private_function_name(
        Identifier(f"write_{enumeration.name}_property")
    )
    to_string_name = golang_naming.function_name(
        Identifier(f"{enumeration.name}_to_string")
    )

    return Stripped(
        f"""\
// Write the `value` of a property as string representation
// of [aastypes.{enum_name}]
// enclosed in an XML element.
//
// Do not flush.
func {function_name}(
{I}encoder *xml.Encoder,
{I}local string,
{I}value aastypes.{enum_name},
) (err error) {{
{I}text, ok := aasstringification.{to_string_name}(
{II}value,
{I})
{I}if !ok {{
{II}err = newSerializationError(
{III}fmt.Sprintf(
{IIII}"Unexpected literal of {enum_name}: %v",
{IIII}value,
{III}),
{II})
{II}return
{I}}}

{I}err = writeStringProperty(encoder, local, text)
{I}return
}}"""
    )


_WRITE_FUNCTION_BY_PRIMITIVE_TYPE = {
    intermediate.PrimitiveType.BOOL: "writeBooleanProperty",
    intermediate.PrimitiveType.INT: "writeLongProperty",
    intermediate.PrimitiveType.FLOAT: "writeDoubleProperty",
    intermediate.PrimitiveType.STR: "writeStringProperty",
    intermediate.PrimitiveType.BYTEARRAY: "writeBytesProperty",
}
assert all(
    literal in _WRITE_FUNCTION_BY_PRIMITIVE_TYPE
    for literal in intermediate.PrimitiveType
)


def _generate_snippet_to_serialize_property(prop: intermediate.Property) -> Stripped:
    blocks = []  # type: List[Stripped]

    local_literal = golang_common.string_literal(naming.xml_property(prop.name))

    segment_name_literal = golang_common.string_literal(
        f"{golang_naming.getter_name(prop.name)}()"
    )

    type_anno = intermediate.beneath_optional(prop.type_annotation)

    getter_name = golang_naming.getter_name(prop.name)
    access_expr = f"that.{getter_name}()"
    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        prop_var = golang_naming.variable_name(Identifier(f"the_{prop.name}"))
        access_expr = prop_var

        blocks.append(
            Stripped(
                f"""\
{prop_var} := that.{getter_name}()"""
            )
        )

    write_block = None  # type: Optional[Stripped]

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation) or (
        isinstance(type_anno, intermediate.OurTypeAnnotation)
        and isinstance(
            type_anno.our_type,
            (intermediate.ConstrainedPrimitive, intermediate.Enumeration),
        )
    ):
        primitive_type = intermediate.try_primitive_type(type_anno)

        if primitive_type is not None:
            write_function = _WRITE_FUNCTION_BY_PRIMITIVE_TYPE[primitive_type]
        else:
            assert isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
                type_anno.our_type, intermediate.Enumeration
            )
            write_function = golang_naming.private_function_name(
                Identifier(f"write_{type_anno.our_type.name}_property")
            )

        pointer = golang_pointering.is_pointer_type(prop.type_annotation)

        if_err_nil_prepend_name_if_serialization_error = Stripped(
            f"""\
if err != nil {{
{I}if seriaErr, ok := err.(*SerializationError); ok {{
{II}seriaErr.Path.PrependName(
{III}&aasreporting.NameSegment{{
{IIII}Name: {segment_name_literal},
{III}}},
{II})
{I}}}
{I}return
}}"""
        )

        if pointer:
            write_block = Stripped(
                f"""\
err = {write_function}(
{I}encoder,
{I}{local_literal},
{I}*{access_expr},
)
{if_err_nil_prepend_name_if_serialization_error}"""
            )
        else:
            write_block = Stripped(
                f"""\
err = {write_function}(
{I}encoder,
{I}{local_literal},
{I}{access_expr},
)
{if_err_nil_prepend_name_if_serialization_error}"""
            )

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        our_type = type_anno.our_type

        if isinstance(our_type, intermediate.Enumeration):
            raise AssertionError("Must have been handled before")

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            raise AssertionError("Must have been handled before")

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if (
                isinstance(our_type, intermediate.ConcreteClass)
                and len(our_type.concrete_descendants) == 0
            ):
                write_as_sequence_name = golang_naming.private_function_name(
                    Identifier(f"write_{our_type.name}_as_sequence")
                )
                write_block = Stripped(
                    f"""\
err = writeStartElement(
{I}encoder,
{I}{local_literal},
{I}false,
)
if err != nil {{
{I}return
}}
err = {write_as_sequence_name}(
{I}encoder,
{I}{access_expr},
)
if err != nil {{
{I}if seriaErr, ok := err.(*SerializationError); ok {{
{II}seriaErr.Path.PrependName(
{III}&aasreporting.NameSegment{{
{IIII}Name: {segment_name_literal},
{III}}},
{II})
{I}}}
{I}return
}}
err = writeEndElement(
{I}encoder,
{I}{local_literal},
{I}false,
)
if err != nil {{
{I}return
}}"""
                )
            else:
                write_block = Stripped(
                    f"""\
err = writeStartElement(
{I}encoder,
{I}{local_literal},
false,
)
if err != nil {{
{I}return
}}
err = Marshal(
{I}encoder,
{I}{access_expr},
{I}false,
)
if err != nil {{
{I}if seriaErr, ok := err.(*SerializationError); ok {{
{II}seriaErr.Path.PrependName(
{III}&aasreporting.NameSegment{{
{IIII}Name: {segment_name_literal},
{III}}},
{II})
{I}}}
{I}return
}}
err = writeEndElement(
{I}encoder,
{I}{local_literal},
false,
)
if err != nil {{
{I}return
}}"""
                )

        else:
            assert_never(our_type)

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert isinstance(
            type_anno.items, intermediate.OurTypeAnnotation
        ) and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass),
        ), (
            f"NOTE (mristin, 2023-06-20): We expect only lists of classes "
            f"at the moment, but you specified {type_anno}. "
            f"Please contact the developers if you need this feature."
        )

        write_block = Stripped(
            f"""\
err = writeListProperty(
{I}encoder,
{I}{local_literal},
{I}{access_expr},
)
if err != nil {{
{I}if seriaErr, ok := err.(*SerializationError); ok {{
{II}seriaErr.Path.PrependName(
{III}&aasreporting.NameSegment{{
{IIII}Name: {segment_name_literal},
{III}}},
{II})
{I}}}
{I}return
}}"""
        )

    else:
        assert_never(type_anno)

    assert write_block is not None

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        blocks.append(
            Stripped(
                f"""\
if {access_expr} != nil {{
{I}{indent_but_first_line(write_block, I)}
}}"""
            )
        )
    else:
        blocks.append(write_block)

    blocks.append(
        Stripped(
            f"""\
err = encoder.Flush()
if err != nil {{
{I}return err
}}"""
        )
    )

    blocks.insert(0, Stripped(f"// region {getter_name}"))
    blocks.append(Stripped("// endregion"))

    return Stripped("\n\n".join(blocks))


def _generate_write_as_sequence(cls: intermediate.ConcreteClass) -> Stripped:
    function_name = golang_naming.private_function_name(
        Identifier(f"write_{cls.name}_as_sequence")
    )

    interface_name = golang_naming.interface_name(cls.name)

    prop_blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        prop_blocks.append(_generate_snippet_to_serialize_property(prop=prop))

    if len(prop_blocks) == 0:
        prop_blocks.append(Stripped("// Intentionally empty."))

    prop_blocks_joined = "\n\n".join(prop_blocks)

    return Stripped(
        f"""\
// Serialize the instance
// of [aastypes.{interface_name}]
// as a sequence of properties, each represented as an XML element.
//
// The XML namespace is expected to be set in the one of the parent elements
// enclosing the sequence.
//
// Flush at the end element of each property.
func {function_name}(
{I}encoder *xml.Encoder,
{I}that aastypes.{interface_name},
) (err error) {{
{I}{indent_but_first_line(prop_blocks_joined, I)}

{I}return
}}"""
    )


def _generate_write_for(cls: intermediate.ConcreteClass) -> Stripped:
    interface_name = golang_naming.interface_name(cls.name)

    if len(cls.concrete_descendants) == 0:
        function_name = golang_naming.private_function_name(
            Identifier(f"write_{cls.name}")
        )
        doc_comment = Stripped(
            f"""\
// Serialize the instance of [aastypes.{interface_name}]
// enclosed in an XML element which represents the model type.
//
// If `withNamespace` is set, the `xmlns` attribute is set in the outer XML element.
//
// Flush once the closing end element has been written."""
        )
    else:
        model_type_literal = golang_naming.enum_literal_name(
            enumeration_name=Identifier("Model_type"), literal_name=cls.name
        )

        doc_comment = Stripped(
            f"""\
// Serialize the instance of [aastypes.{interface_name}]
// enclosed in an XML element which represents the model type.
//
// Do not dispatch on the runtime model type, *i.e.*, assume that the runtime model type
// is exactly [aastypes.{model_type_literal}]. If you need dispatch,
// call [Marshal].
//
// If `withNamespace` is set, the `xmlns` attribute is set in the outer XML element.
//
// Flush once the closing end element has been written."""
        )

        function_name = golang_naming.private_function_name(
            Identifier(f"write_{cls.name}_without_dispatch")
        )

    xml_class_name_literal = golang_common.string_literal(
        naming.xml_class_name(cls.name)
    )

    write_as_sequence_name = golang_naming.private_function_name(
        Identifier(f"write_{cls.name}_as_sequence")
    )

    return Stripped(
        f"""\
{doc_comment}
func {function_name}(
{I}encoder *xml.Encoder,
{I}that aastypes.{interface_name},
{I}withNamespace bool,
) (err error) {{
{I}local := {xml_class_name_literal}
{I}
{I}err = writeStartElement(
{II}encoder,
{II}local,
{II}withNamespace,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}err = {write_as_sequence_name}(
{II}encoder,
{II}that,
{I})
{I}if err != nil {{
{II}return
{I}}}
{I}
{I}err = writeEndElement(
{II}encoder,
{II}local,
{II}withNamespace,
{I})
{I}if err != nil {{
{II}return
{I}}}

{I}err = encoder.Flush()
{I}return
}}"""
    )


def _generate_marshal(symbol_table: intermediate.SymbolTable) -> Stripped:
    case_blocks = []  # type: List[Stripped]
    for cls in symbol_table.concrete_classes:
        model_type_literal = golang_naming.enum_literal_name(
            enumeration_name=Identifier("Model_type"), literal_name=cls.name
        )

        interface_name = golang_naming.interface_name(cls.name)

        if len(cls.concrete_descendants) == 0:
            write_function = golang_naming.private_function_name(
                Identifier(f"write_{cls.name}")
            )
        else:
            write_function = golang_naming.private_function_name(
                Identifier(f"write_{cls.name}_without_dispatch")
            )

        case_blocks.append(
            Stripped(
                f"""\
case aastypes.{model_type_literal}:
{I}err = {write_function}(
{II}encoder,
{II}that.(aastypes.{interface_name}),
{II}withNamespace,
{I})"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}err = newSerializationError(
{II}fmt.Sprintf(
{III}"Unexpected model type: %v",
{III}that.ModelType(),
{II}),
{I})"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)
    model_type_getter = golang_naming.getter_name(Identifier("model_type"))

    return Stripped(
        f"""\
// Serialize `that` instance as an XML element.
//
// If `withNamespace` is set, the `xmlns` attribute is set in the XML element
// to [Namespace].
func Marshal(
{I}encoder *xml.Encoder,
{I}that aastypes.IClass,
{I}withNamespace bool,
) (err error) {{
{I}switch that.{model_type_getter}() {{
{I}{indent_but_first_line(case_blocks_joined, I)}
{I}}}
{I}return
}}"""
    )


# endregion

# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
    repo_url: Stripped,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the Golang code for the general de/serialization."""
    aastypes_url_literal = golang_common.string_literal(f"{repo_url}/types")

    aasreporting_url_literal = golang_common.string_literal(f"{repo_url}/reporting")

    aasstringification_url_literal = golang_common.string_literal(
        f"{repo_url}/stringification"
    )

    aasverification_url_literal = golang_common.string_literal(
        f"{repo_url}/verification"
    )

    namespace_literal = golang_common.string_literal(
        symbol_table.meta_model.xml_namespace
    )

    blocks = [
        Stripped(
            """\
// Package xmlization de/serializes model instances to and from XML.
//
// To de-serialize, call one of the `Unmarshal*` functions.
//
// To serialize, call the [Marshal] function.
package xmlization"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}b64 "encoding/base64"
{I}"encoding/xml"
{I}"fmt"
{I}"io"
{I}"math"
{I}"strconv"
{I}"strings"
{I}"unicode"
{I}aasreporting {aasreporting_url_literal}
{I}aasstringification {aasstringification_url_literal}
{I}aastypes {aastypes_url_literal}
{I}aasverification {aasverification_url_literal}
)"""
        ),
        Stripped("// region De-serialization"),
    ]

    blocks.extend(_generate_deserialization_error_and_its_methods())

    blocks.extend(
        [
            Stripped(
                """\
// This is class for a sentinel token to signal the end-of-file.
type eof struct{}"""
            ),
            _generate_is_whitespace(),
            _generate_read_next(),
            _generate_skip_empty_text_whitespace_and_comments(),
            _generate_read_text(),
            _generate_read_text_as_boolean(),
            _generate_read_text_as_long(),
            _generate_read_text_as_double(),
            _generate_read_text_as_base64_encoded_bytes(),
            Stripped(
                f"""\
const Namespace = {namespace_literal}"""
            ),
            _generate_check_start_element(),
            _generate_extract_local_name_from_start_element(),
            _generate_parse_as_start_element_and_extract_local_name(),
            _generate_check_end_element(),
            _generate_read_list(),
        ]
    )

    errors = []  # type: List[Error]

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            blocks.append(_generate_read_text_as_enumeration(enumeration=our_type))

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            pass
        elif isinstance(our_type, intermediate.AbstractClass):
            blocks.append(_generate_read_with_lookahead_with_dispatch(cls=our_type))

        elif isinstance(our_type, intermediate.ConcreteClass):
            if our_type.is_implementation_specific:
                implementation_key = specific_implementations.ImplementationKey(
                    f"Xmlization/read_{our_type.name}_as_sequence.go"
                )

                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            our_type.parsed.node,
                            f"The xmlization snippet is missing "
                            f"for the implementation-specific "
                            f"class {our_type.name}: {implementation_key}",
                        )
                    )
                    continue
            else:
                blocks.append(_generate_read_as_sequence(cls=our_type))

            if len(our_type.concrete_descendants) > 0:
                blocks.append(_generate_read_with_lookahead_with_dispatch(cls=our_type))
            else:
                blocks.append(
                    _generate_read_with_lookahead_without_dispatch(cls=our_type)
                )
        else:
            assert_never(our_type)

    blocks.append(_generate_unmarshal(symbol_table=symbol_table))

    blocks.append(Stripped("// endregion"))

    blocks.append(Stripped("// region Serialization"))

    blocks.extend(_generate_serialization_error())

    blocks.extend(
        [
            _generate_write_start_element(),
            _generate_write_end_element(),
            _generate_write_text(),
            _generate_write_boolean_property(),
            _generate_write_long_property(),
            _generate_write_double_property(),
            _generate_write_string_property(),
            _generate_write_bytes_property(),
            _generate_write_list_as_sequence(),
        ]
    )

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            blocks.append(_generate_write_enumeration_property(enumeration=our_type))

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2023-06-18):
            # We will serialize constrained primitives as primitives.
            pass

        elif isinstance(our_type, intermediate.AbstractClass):
            # NOTE (mristin, 2023-06-18):
            # We will use general ``write`` function.
            pass

        elif isinstance(our_type, intermediate.ConcreteClass):
            if our_type.is_implementation_specific:
                implementation_key = specific_implementations.ImplementationKey(
                    f"Xmlization/write_{our_type.name}_as_sequence.go"
                )

                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            our_type.parsed.node,
                            f"The xmlization snippet is missing "
                            f"for the implementation-specific "
                            f"class {our_type.name}: {implementation_key}",
                        )
                    )
                    continue
            else:
                blocks.append(_generate_write_as_sequence(cls=our_type))

            blocks.append(_generate_write_for(cls=our_type))
        else:
            assert_never(our_type)

    blocks.append(_generate_marshal(symbol_table=symbol_table))

    blocks.append(Stripped("// endregion"))

    if len(errors) > 0:
        return None, errors

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None



"""Generate the test code for the JSON de/serialization of concrete classes."""

from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> str:
    """
    Generate the test code for the JSON de/serialization of concrete classes.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    blocks = [
        Stripped(
            f"""\
private static void AssertSerializeDeserializeEqualsOriginal(
{I}Nodes.JsonNode originalNode, Aas.IClass instance, string path)
{{
{I}Nodes.JsonObject? serialized = null;
{I}try
{I}{{
{II}serialized = Aas.Jsonization.Serialize.ToJsonObject(instance);
{I}}}
{I}catch (System.Exception exception)
{I}{{
{II}Assert.Fail(
{III}"Expected no exception upon serialization of an instance " +
{III}$"de-serialized from {{path}}, but got: {{exception}}"
{II});
{I}}}

{I}if (serialized == null)
{I}{{
{II}Assert.Fail(
{III}$"Unexpected null serialization of an instance from {{path}}"
{II});
{I}}}
{I}else
{I}{{
{II}Aas.Tests.CommonJson.CheckJsonNodesEqual(
{III}originalNode,
{III}serialized,
{III}out Reporting.Error? inequalityError);
{II}if (inequalityError != null)
{II}{{
{III}Assert.Fail(
{IIII}$"The original JSON from {{path}} is unequal the serialized JSON: " +
{IIII}$"{{Reporting.GenerateJsonPath(inequalityError.PathSegments)}}: " +
{IIII}inequalityError.Cause
{III});
{II}}}
{I}}}
}}"""
        ),
        Stripped(
            f"""\
private static void AssertEqualsExpectedOrRerecordDeserializationException(
{I}Aas.Jsonization.Exception? exception,
{I}string path)
{{
{I}if (exception == null)
{I}{{
{II}Assert.Fail(
{III}$"Expected a Jsonization exception when de-serializing {{path}}, but got none."
{II});
{I}}}
{I}else
{I}{{
{II}string exceptionPath = path + ".exception";
{II}string got = exception.Message;
{II}if (Aas.Tests.Common.RecordMode)
{II}{{
{III}System.IO.File.WriteAllText(exceptionPath, got);
{II}}}
{II}else
{II}{{
{III}if (!System.IO.File.Exists(exceptionPath))
{III}{{
{IIII}throw new System.IO.FileNotFoundException(
{IIIII}$"The file with the recorded exception does not exist: {{exceptionPath}}; " +
{IIIII}"maybe you want to set the environment " +
{IIIII}$"variable {{Aas.Tests.Common.RecordModeEnvironmentVariableName}}?");
{III}}}

{III}string expected = System.IO.File.ReadAllText(exceptionPath);
{III}Assert.AreEqual(
{IIII}expected.Replace("\\r\\n", "\\n"),
{IIII}got.Replace("\\r\\n", "\\n"),
{IIII}$"The expected exception does not match the actual one for the file {{path}}");
{II}}}
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_csharp = csharp_naming.class_name(concrete_cls.name)
        cls_name_json = naming.json_model_type(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_{cls_name_csharp}_ok()
{{
{I}var paths = Directory.GetFiles(
{II}Path.Combine(
{III}Aas.Tests.Common.TestDataDir,
{III}"Json",
{III}"Expected",
{III}{csharp_common.string_literal(cls_name_json)}
{II}),
{II}"*.json",
{II}System.IO.SearchOption.AllDirectories).ToList();
{I}paths.Sort();

{I}foreach (var path in paths)
{I}{{
{II}var node = Aas.Tests.CommonJson.ReadFromFile(path);

{II}var instance = Aas.Jsonization.Deserialize.{cls_name_csharp}From(
{III}node);

{II}var errors = Aas.Verification.Verify(instance).ToList();
{II}Aas.Tests.Common.AssertNoVerificationErrors(errors, path);

{II}AssertSerializeDeserializeEqualsOriginal(
{III}node, instance, path);
{I}}}
}}  // public void Test_{cls_name_csharp}_ok"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_{cls_name_csharp}_deserialization_from_non_object_fail()
{{
{I}var node = Nodes.JsonValue.Create("INVALID")
{II}?? throw new System.InvalidOperationException(
{III}"Unexpected failure of the node creation");

{I}Aas.Jsonization.Exception? exception = null;
{I}try
{I}{{
{II}var _ = Aas.Jsonization.Deserialize.{cls_name_csharp}From(
{III}node);
{I}}}
{I}catch (Aas.Jsonization.Exception observedException)
{I}{{
{II}exception = observedException;
{I}}}

{I}if (exception == null)
{I}{{
{II}throw new AssertionException("Expected an exception, but got none");
{I}}}

{I}if (!exception.Message.StartsWith("Expected a JsonObject, but got "))
{I}{{
{II}throw new AssertionException(
{III}$"Unexpected exception message: {{exception.Message}}");
{I}}}
}}  // public void Test_{cls_name_csharp}_deserialization_from_non_object_fail"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_{cls_name_csharp}_deserialization_fail()
{{
{I}foreach (
{II}string causeDir in
{II}Directory.GetDirectories(
{III}Path.Combine(
{IIII}Aas.Tests.Common.TestDataDir,
{IIII}"Json",
{IIII}"Unexpected",
{IIII}"Unserializable"
{III})
{II})
{I})
{I}{{
{II}string clsDir = Path.Combine(
{III}causeDir,
{III}{csharp_common.string_literal(cls_name_json)}
{II});

{II}if (!Directory.Exists(clsDir))
{II}{{
{III}// No examples of {cls_name_csharp} for the failure cause.
{III}continue;
{II}}}

{II}var paths = Directory.GetFiles(
{III}clsDir,
{III}"*.json",
{III}System.IO.SearchOption.AllDirectories).ToList();
{II}paths.Sort();

{II}foreach (var path in paths)
{II}{{
{III}var node = Aas.Tests.CommonJson.ReadFromFile(path);

{III}Aas.Jsonization.Exception? exception = null;
{III}try
{III}{{
{IIII}var _ = Aas.Jsonization.Deserialize.{cls_name_csharp}From(
{IIIII}node);
{III}}}
{III}catch (Aas.Jsonization.Exception observedException)
{III}{{
{IIII}exception = observedException;
{III}}}

{III}AssertEqualsExpectedOrRerecordDeserializationException(
{IIII}exception, path);
{II}}}
{I}}}
}}  // public void Test_{cls_name_csharp}_deserialization_fail"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_{cls_name_csharp}_verification_fail()
{{
{I}foreach (
{II}string causeDir in
{II}Directory.GetDirectories(
{III}Path.Combine(
{IIII}Aas.Tests.Common.TestDataDir,
{IIII}"Json",
{IIII}"Unexpected",
{IIII}"Invalid"
{III})
{II})
{I})
{I}{{
{II}string clsDir = Path.Combine(
{III}causeDir,
{III}{csharp_common.string_literal(cls_name_json)}
{II});

{II}if (!Directory.Exists(clsDir))
{II}{{
{III}// No examples of {cls_name_csharp} for the failure cause.
{III}continue;
{II}}}

{II}var paths = Directory.GetFiles(
{III}clsDir,
{III}"*.json",
{III}System.IO.SearchOption.AllDirectories).ToList();
{II}paths.Sort();

{II}foreach (var path in paths)
{II}{{
{III}var node = Aas.Tests.CommonJson.ReadFromFile(path);

{III}var instance = Aas.Jsonization.Deserialize.{cls_name_csharp}From(
{IIII}node);

{III}var errors = Aas.Verification.Verify(instance).ToList();
{III}Aas.Tests.Common.AssertEqualsExpectedOrRerecordVerificationErrors(
{IIII}errors, path);
{II}}}
{I}}}
}}  // public void Test_{cls_name_csharp}_verification_fail"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}

using Aas = {namespace};  // renamed

using Directory = System.IO.Directory;
using Nodes = System.Text.Json.Nodes;
using Path = System.IO.Path;

using System.Linq;  // can't alias
using NUnit.Framework; // can't alias

namespace {namespace}.Tests
{{
{I}public class TestJsonizationOfConcreteClasses
{I}{{
{II}{indent_but_first_line(blocks_joined, II)}
{I}}}  // class TestJsonizationOfConcreteClasses
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate the shared code used across the jsonization unit tests."""
import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, Identifier, indent_but_first_line
from aas_core_codegen.python import common as python_common, naming as python_naming
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
)


def _generate_model_type_to_from_jsonable(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the map from model type to the ``*_from_jsonable`` function."""
    items = []  # type: List[Stripped]

    for cls in sorted(symbol_table.concrete_classes, key=lambda a_cls: a_cls.name):
        model_type = naming.json_model_type(cls.name)
        from_jsonable = python_naming.function_name(
            Identifier(f"{cls.name}_from_jsonable")
        )

        items.append(
            Stripped(f"{repr(model_type)}:\n{I}aas_jsonization.{from_jsonable}")
        )

    items_joined = ",\n".join(items)

    return Stripped(
        f"""\
_MODEL_TYPE_TO_FROM_JSONABLE: Mapping[
{I}str,
{I}Callable[[aas_jsonization.Jsonable], aas_types.Class]
] = {{
{I}{indent_but_first_line(items_joined, I)}
}}"""
    )


def _generate_model_type_to_class(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the map from the concrete class to model type."""
    items = []  # type: List[Stripped]

    for cls in sorted(symbol_table.concrete_classes, key=lambda a_cls: a_cls.name):
        model_type = naming.json_model_type(cls.name)
        python_cls_name = python_naming.class_name(cls.name)

        items.append(Stripped(f"{repr(model_type)}:\n{I}aas_types.{python_cls_name}"))

    items_joined = ",\n".join(items)

    return Stripped(
        f"""\
_MODEL_TYPE_TO_CLASS: Mapping[
{I}str,
{I}Type[aas_types.Class]
] = {{
{I}{indent_but_first_line(items_joined, I)}
}}"""
    )


@ensure(
    lambda result: result.endswith("\n"),
    "Trailing newline mandatory for valid end-of-files",
)
def generate(
    symbol_table: intermediate.SymbolTable,
    qualified_module_name: python_common.QualifiedModuleName,
) -> str:
    """
    Generate the shared code used across the jsonization unit tests.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    blocks = [
        Stripped(
            '"""Provide common functionality for JSON de/serialization in tests."""',
        ),
        python_common.WARNING,
        Stripped(
            """\
# pylint: disable=missing-docstring"""
        ),
        Stripped(
            """\
import collections.abc
import json
import pathlib
import sys
from typing import Callable, Iterator, Mapping, Type"""
        ),
        Stripped(
            f"""\
if sys.version_info >= (3, 8):
{I}from typing import Final
else:
{I}from typing_extensions import Final"""
        ),
        Stripped(
            f"""\
import {qualified_module_name}.common as aas_common
import {qualified_module_name}.jsonization as aas_jsonization
import {qualified_module_name}.types as aas_types"""
        ),
        Stripped(
            f'''\
class Difference:
{I}"""Represent a single difference between two JSON-ables."""

{I}#: Human-readable description of the difference
{I}message: Final[str]

{I}#: Path in the expected JSON-able value which is different from
{I}#: the obtained JSON-able value
{I}path: Final[aas_jsonization.Path]

{I}def __init__(self, message: str) -> None:
{II}"""Initialize with the given message and empty path."""
{II}self.message = message
{II}self.path = aas_jsonization.Path()

{I}def __str__(self) -> str:
{II}return f"#{{self.path}}: {{self.message}}"'''
        ),
        Stripped(
            f'''\
def check_equal(
{I}expected: aas_jsonization.Jsonable,
{I}got: aas_jsonization.Jsonable,
) -> Iterator[Difference]:
{I}"""
{I}Compare recursively two JSON-able values for equality.

{I}:param expected: expected JSON-able value
{I}:param got: obtained JSON-able value
{I}:yield: differences
{I}"""
{I}if isinstance(expected, (bool, int, float, str, bytes)):
{II}if type(expected) != type(got):  # pylint: disable=unidiomatic-typecheck
{III}yield Difference(f"Expected {{type(expected)}}, but got {{type(got)}}")

{II}if expected != got:
{III}yield Difference(f"Expected {{expected!r}}, but got {{got!r}}")
{I}elif isinstance(expected, collections.abc.Sequence):
{II}if not isinstance(got, collections.abc.Sequence):
{III}yield Difference(f"Expected a sequence, but got {{type(got)}}")
{II}else:
{III}if len(expected) != len(got):
{IIII}yield Difference(
{IIIII}f"Expected a sequence of {{len(expected)}} item(s), "
{IIIII}f"but got {{len(got)}} item(s)"
{IIII})

{III}for i, (expected_item, got_item) in enumerate(zip(expected, got)):
{IIII}for difference in check_equal(expected_item, got_item):
{IIIII}difference.path._prepend(aas_jsonization.IndexSegment(expected, i))
{IIIII}yield difference

{I}elif isinstance(expected, collections.abc.Mapping):
{II}if not isinstance(got, collections.abc.Mapping):
{III}yield Difference(f"Expected a mapping, but got {{type(got)}}")
{II}else:
{III}if not all(isinstance(key, str) for key in expected.keys()):
{IIII}raise ValueError(
{IIIII}f"Expected all keys in the expected JSON-able value to be strings, "
{IIIII}f"but got: {{list(expected.keys())}}"
{IIII})

{III}if not all(isinstance(key, str) for key in got.keys()):
{IIII}yield Difference(
{IIIII}f"Expected all keys in a mapping to be strings, "
{IIIII}f"but got: {{list(got.keys())}}"
{IIII})

{III}expected_key_set = set(expected.keys())
{III}got_key_set = set(got.keys())

{III}expected_got_diff = expected_key_set.difference(got_key_set)
{III}if expected_got_diff:
{IIII}yield Difference(f"Expected key(s) {{sorted(expected_got_diff)}} missing")

{III}got_expected_diff = got_key_set.difference(expected_key_set)
{III}if got_expected_diff:
{IIII}yield Difference(f"Unexpected key(s) {{sorted(got_expected_diff)}}")

{III}for key, expected_value in expected.items():
{IIII}got_value = got[key]

{IIII}for difference in check_equal(expected_value, got_value):
{IIIII}difference.path._prepend(
{IIIIII}aas_jsonization.PropertySegment(expected, key)
{IIIII})
{IIIII}yield difference
{I}else:
{II}aas_common.assert_never(expected)'''
        ),
        _generate_model_type_to_from_jsonable(symbol_table=symbol_table),
        _generate_model_type_to_class(symbol_table=symbol_table),
        Stripped(
            f'''\
def must_load(
{I}path: pathlib.Path,
{I}model_type: str
) -> aas_types.Class:
{I}"""
{I}Load an instance from ``path``.

{I}The class of the instance must correspond to the ``model_type``.
{I}"""
{I}try:
{II}jsonable = json.loads(
{III}path.read_text(encoding='utf-8')
{II})
{I}except Exception as exception:
{II}raise RuntimeError(f"Failed to read from {{path}}") from exception

{I}from_jsonable = _MODEL_TYPE_TO_FROM_JSONABLE.get(model_type, None)
{I}if from_jsonable is None:
{II}raise AssertionError(
{III}"The model type is not mapped to "
{III}f"the from_jsonable function: {{model_type!r}}"
{II})

{I}try:
{II}instance = from_jsonable(jsonable)
{I}except Exception as exception:
{II}raise RuntimeError(
{III}"Failed to parse the instance "
{III}f"with model type {{model_type}} from {{path}}"
{II}) from exception

{I}cls = _MODEL_TYPE_TO_CLASS.get(model_type, None)
{I}if cls is None:
{II}raise AssertionError(
{III}"The model type is not mapped to the class: {{model_type!r}}"
{II})

{I}if not isinstance(instance, cls):
{II}raise RuntimeError(
{III}f"Expected an instance with model type {{model_type!r}}, "
{III}f"corresponding to class {{cls}}, "
{III}f"but got an instance {{instance}}"
{II})

{I}return instance'''
        ),
        python_common.WARNING,
    ]

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue()



"""Provide common functions shared among different C# code generation modules."""
import re
from typing import List, cast, Optional

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, assert_never
from aas_core_codegen.csharp import naming as csharp_naming


@ensure(lambda result: result.startswith('"'))
@ensure(lambda result: result.endswith('"'))
def string_literal(text: str) -> Stripped:
    """Generate a C# string literal from the ``text``."""
    escaped = []  # type: List[str]

    for character in text:
        if character == "\a":
            escaped.append("\\a")
        elif character == "\b":
            escaped.append("\\b")
        elif character == "\f":
            escaped.append("\\f")
        elif character == "\n":
            escaped.append("\\n")
        elif character == "\r":
            escaped.append("\\r")
        elif character == "\t":
            escaped.append("\\t")
        elif character == "\v":
            escaped.append("\\v")
        elif character == '"':
            escaped.append('\\"')
        elif character == "\\":
            escaped.append("\\\\")
        else:
            escaped.append(character)

    return Stripped('"{}"'.format("".join(escaped)))


def needs_escaping(text: str) -> bool:
    """Check whether the ``text`` contains a character that needs escaping."""
    for character in text:
        if character == "\a":
            return True
        elif character == "\b":
            return True
        elif character == "\f":
            return True
        elif character == "\n":
            return True
        elif character == "\r":
            return True
        elif character == "\t":
            return True
        elif character == "\v":
            return True
        elif character == '"':
            return True
        elif character == "\\":
            return True
        else:
            pass

    return False


PRIMITIVE_TYPE_MAP = {
    intermediate.PrimitiveType.BOOL: Stripped("bool"),
    intermediate.PrimitiveType.INT: Stripped("long"),
    intermediate.PrimitiveType.FLOAT: Stripped("double"),
    intermediate.PrimitiveType.STR: Stripped("string"),
    intermediate.PrimitiveType.BYTEARRAY: Stripped("byte[]"),
}


def _assert_all_primitive_types_are_mapped() -> None:
    """Assert that we have explicitly mapped all the primitive types to C#."""
    all_primitive_literals = set(literal.value for literal in PRIMITIVE_TYPE_MAP)

    mapped_primitive_literals = set(
        literal.value for literal in intermediate.PrimitiveType
    )

    all_diff = all_primitive_literals.difference(mapped_primitive_literals)
    mapped_diff = mapped_primitive_literals.difference(all_primitive_literals)

    messages = []  # type: List[str]
    if len(mapped_diff) > 0:
        messages.append(
            f"More primitive maps are mapped than there were defined "
            f"in the ``intermediate._types``: {sorted(mapped_diff)}"
        )

    if len(all_diff) > 0:
        messages.append(
            f"One or more primitive types in the ``intermediate._types`` were not "
            f"mapped in PRIMITIVE_TYPE_MAP: {sorted(all_diff)}"
        )

    if len(messages) > 0:
        raise AssertionError("\n\n".join(messages))


_assert_all_primitive_types_are_mapped()


# fmt: off
@require(
    lambda our_type_qualifier:
    not (our_type_qualifier is not None)
    or not our_type_qualifier.endswith('.')
)
# fmt: on
def generate_type(
    type_annotation: intermediate.TypeAnnotationUnion,
    our_type_qualifier: Optional[Stripped] = None,
) -> Stripped:
    """
    Generate the C# type for the given type annotation.

    ``our_type_prefix`` is appended to all our types, if specified.
    """
    our_type_prefix = "" if our_type_qualifier is None else f"{our_type_qualifier}."
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return PRIMITIVE_TYPE_MAP[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type

        if isinstance(our_type, intermediate.Enumeration):
            return Stripped(
                our_type_prefix + csharp_naming.enum_name(type_annotation.our_type.name)
            )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return PRIMITIVE_TYPE_MAP[our_type.constrainee]

        elif isinstance(our_type, intermediate.Class):
            # NOTE (mristin, 2023-02-08):
            # We want to allow custom enhancements and wrappings around
            # our model classes. Therefore, we always operate over C# interfaces
            # instead of concrete classes, even if the class is a concrete one and
            # has no concrete descendants.

            return Stripped(
                our_type_prefix + csharp_naming.interface_name(our_type.name)
            )

    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        item_type = generate_type(
            type_annotation=type_annotation.items, our_type_qualifier=our_type_qualifier
        )

        return Stripped(f"List<{item_type}>")

    elif isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        value = generate_type(
            type_annotation=type_annotation.value, our_type_qualifier=our_type_qualifier
        )
        return Stripped(f"{value}?")

    else:
        assert_never(type_annotation)

    raise AssertionError("Should not have gotten here")


INDENT = "    "
INDENT2 = INDENT * 2
INDENT3 = INDENT * 3
INDENT4 = INDENT * 4
INDENT5 = INDENT * 5
INDENT6 = INDENT * 6
INDENT7 = INDENT * 7
INDENT8 = INDENT * 8

# noinspection RegExpSimplifiable
NAMESPACE_IDENTIFIER_RE = re.compile(
    r"[a-zA-Z_][a-zA-Z_0-9]*(\.[a-zA-Z_][a-zA-Z_0-9]*)*"
)


class NamespaceIdentifier(str):
    """Capture a namespace identifier."""

    @require(lambda identifier: NAMESPACE_IDENTIFIER_RE.fullmatch(identifier))
    def __new__(cls, identifier: str) -> "NamespaceIdentifier":
        return cast(NamespaceIdentifier, identifier)


WARNING = Stripped(
    """\
/*
 * This code has been automatically generated by aas-core-codegen.
 * Do NOT edit or append.
 */"""
)


# fmt: off
@ensure(
    lambda namespace, result:
    not (namespace != "Aas") or len(result) == 1,
    "Exactly one block of stripped text to be appended to the list of using directives "
    "if this using directive is necessary"
)
@ensure(
    lambda namespace, result:
    not (namespace == "Aas") or len(result) == 0,
    "Empty list if no directive is necessary"
)
# fmt: on
def generate_using_aas_directive_if_necessary(
    namespace: NamespaceIdentifier,
) -> List[Stripped]:
    """Generate the using directive if the namespace does not equal ``Aas``."""
    if namespace == "Aas":
        return []

    if namespace.endswith(".Aas"):
        return [Stripped(f"using Aas = {namespace};")]

    return [Stripped(f"using Aas = {namespace};  // renamed")]



"""Render descriptions to Java documentation comments."""
import dataclasses
import html
import io
import textwrap
import urllib.parse
from typing import (
    Tuple,
    Optional,
    List,
    Sequence,
    Final,
    Union,
)

import docutils.nodes
import docutils.utils
from icontract import require

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    Error,
    assert_never,
    Identifier,
)
from aas_core_codegen.intermediate import (
    doc as intermediate_doc,
)
from aas_core_codegen.java import (
    common as java_common,
    naming as java_naming,
)


class Context:
    """Structure the context of the rendering of a description."""

    #: Fully qualified dot-path to the base package
    root_package: Final[java_common.PackageIdentifier]

    #: Scope in which the description resides.
    #:
    #: This narrows down the outer scope of the description, while the description
    #: does not necessarily describe a class, but can also describe a method,
    #: a property, an enumeration literal *etc.*
    cls_or_enum: Final[
        Optional[Union[intermediate.ClassUnion, intermediate.Enumeration]]
    ]

    def __init__(
        self,
        package: java_common.PackageIdentifier,
        cls_or_enum: Optional[Union[intermediate.ClassUnion, intermediate.Enumeration]],
    ) -> None:
        """Initialize with the given values."""
        self.root_package = package
        self.cls_or_enum = cls_or_enum


# NOTE (mristin, 2024-03-28):
# The Javadoc markup is a bit peculiar. In many cases, it requires only the opening
# tag, and tolerates (or even encourages) omission of the closing tag. Consequently,
# this allows us to model the text flow operations as a list of tokens (instead, say,
# a more complex document tree).
#
# At this point, we only care about the text flow (*i.e.*, line breaks, bullets,
# and beginnings of the paragraphs). To keep it all simple, we ignore all the other
# Javadoc tags which do not affect the text flow, and capture them in plain text.

# NOTE (mristin, 2024-03-28):
# We make all token classes as data classes so that the comparisons and
# string representations are automatically generated.


@dataclasses.dataclass
class _TokenText:
    """Model rendered description text, except any text flow operation."""

    content: str


@dataclasses.dataclass
class _TokenULOpen:
    """Capture a beginning of an unordered list."""


@dataclasses.dataclass
class _TokenULClose:
    """Capture the end an unordered list."""


@dataclasses.dataclass
class _TokenLI:
    """Capture a beginning of a list item."""


@dataclasses.dataclass
class _TokenP:
    """Capture a beginning of a paragraph."""


_Token = Union[_TokenText, _TokenULOpen, _TokenULClose, _TokenLI, _TokenP]


class _ElementRenderer(intermediate_doc.DocutilsElementTransformer[List[_Token]]):
    """Render descriptions as a content of a docstring."""

    def __init__(self, context: Context) -> None:
        """Initialize with the given values."""
        self.context = context

    def transform_text(
        self, element: docutils.nodes.Text
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        return [
            _TokenText(
                element.astext()
                .replace("&", "&amp;")
                .replace("<", "&lt;")
                .replace(">", "&gt;")
            )
        ], None

    def transform_reference_to_our_type_in_doc(
        self, element: intermediate_doc.ReferenceToOurType
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        reference: Stripped

        if isinstance(element.our_type, intermediate.Enumeration):
            name = java_naming.enum_name(element.our_type.name)
            reference = Stripped(
                f"{{@link {self.context.root_package}.types.enums.{name}}}"
            )
        elif isinstance(element.our_type, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2024-03-27):
            # We do not generate a class for constrained primitives, but we
            # leave it as class name, as that is what we used for ``verify*`` functions.
            name = java_naming.class_name(element.our_type.name)
            reference = Stripped(f"@{{code {name}}}")
        elif isinstance(element.our_type, intermediate.AbstractClass):
            # NOTE (empwilli, 2023-12-14):
            # We do not generate Java code for abstract classes, so we have to refer
            # to the interface.
            name = java_naming.interface_name(element.our_type.name)
            reference = Stripped(
                f"{{@link {self.context.root_package}.types.model.{name}}}"
            )
        elif isinstance(element.our_type, intermediate.ConcreteClass):
            # NOTE (empwilli, 2023-12-14):
            # Though a concrete class can have multiple descendants and the writer
            # might actually want to refer to the *interface* instead of
            # the concrete class, we do the best effort here and resolve it to the
            # name of the concrete class.
            name = java_naming.class_name(element.our_type.name)
            reference = Stripped(
                f"{{@link {self.context.root_package}.types.impl.{name}}}"
            )
        else:
            assert_never(element.our_type)

        return [_TokenText(reference)], None

    def transform_reference_to_attribute_in_doc(
        self, element: intermediate_doc.ReferenceToAttribute
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        if isinstance(element.reference, intermediate_doc.ReferenceToProperty):
            getter_name = java_naming.getter_name(element.reference.prop.name)

            if element.reference.cls is self.context.cls_or_enum:
                return [_TokenText(f"{{@link #{getter_name}()}}")], None

            base: Stripped
            if isinstance(element.reference.cls, intermediate.AbstractClass):
                # NOTE (empwilli, 2023-12-14):
                # We do not generate Java code for abstract classes, so we have to refer
                # to the interface.
                interface_name = java_naming.interface_name(element.reference.cls.name)
                base = Stripped(f"types.model.{interface_name}")
            elif isinstance(element.reference.cls, intermediate.ConcreteClass):
                # NOTE (empwilli, 2023-12-14):
                # Though a concrete class can have multiple descendants and the writer
                # might actually want to refer to the *interface* instead of
                # the concrete class, we do the best effort here and resolve it to the
                # name of the concrete class.
                cls_name = java_naming.class_name(element.reference.cls.name)
                base = Stripped(f"types.impl.{cls_name}")
            else:
                assert_never(element.reference.cls)

            return [
                _TokenText(
                    f"{{@link {self.context.root_package}.{base}#{getter_name}()}}"
                )
            ], None

        elif isinstance(
            element.reference, intermediate_doc.ReferenceToEnumerationLiteral
        ):
            literal_name = java_naming.enum_literal_name(element.reference.literal.name)

            if element.reference.enumeration is self.context.cls_or_enum:
                return [_TokenText(f"{{@link #{literal_name}}}")], None

            enum_name = java_naming.enum_name(element.reference.enumeration.name)

            return [
                _TokenText(
                    f"{{@link {self.context.root_package}.types.enums"
                    f".{enum_name}#{literal_name}}}"
                )
            ], None
        else:
            assert_never(element.reference)

    def transform_reference_to_argument_in_doc(
        self, element: intermediate_doc.ReferenceToArgument
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        arg_name = java_naming.argument_name(Identifier(element.reference))
        return [_TokenText(f"{{@code {arg_name}}}")], None

    def transform_reference_to_constraint_in_doc(
        self, element: intermediate_doc.ReferenceToConstraint
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        assert isinstance(element.reference, str)

        return [_TokenText(f"Constraint {element.reference}")], None

    def transform_reference_to_constant_in_doc(
        self, element: intermediate_doc.ReferenceToConstant
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        name = java_naming.property_name(element.constant.name)

        return [
            _TokenText(
                f"{{@link {self.context.root_package}.constants.Constants#{name}}}"
            )
        ], None

    def transform_literal(
        self, element: docutils.nodes.literal
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        # NOTE (mristin, 2024-03-27):
        # Theoretically, we could escape the backticks properly here, see
        # https://meta.stackexchange.com/questions/82718/how-do-i-escape-a-backtick-within-in-line-code-in-markdown.
        # However, this is not necessary since our meta-model is written in
        # Python, and escaping backticks in ReST is not well-defined. For now, we simply
        # assume no backticks.
        #
        # See: https://stackoverflow.com/questions/66435475/how-to-escape-the-backtick-character-in-a-rst-file
        text = element.astext()
        assert "`" not in text, (
            "(mristin, 2024-03-27): Theoretically, we could escape the backticks "
            "properly here, see [how to escape backticks in markdown].\n\n"
            "However, this is not necessary as our meta-model is written in Python, "
            "and escaping backticks in ReST is not well-defined, see [how to escape "
            "backticks in ReST]. At this moment, we simply assume no backticks in the "
            "code literals.\n\n"
            "[how to escape backticks in markdown]: https://meta.stackexchange.com/"
            "questions/82718/how-do-i-escape-a-backtick-within-in-line-"
            "code-in-markdown\n"
            "[how to escape backticks in ReST]: https://stackoverflow.com/questions/"
            "66435475/how-to-escape-the-backtick-character-in-a-rst-file"
        )

        escaped = (
            text
            # NOTE (mristin, 2024-03-27):
            # See: https://stackoverflow.com/questions/2290757/how-can-you-escape-the-character-in-javadoc
            .replace("@", "&#064;")
        )
        return [_TokenText(f"{{@code {escaped}}}")], None

    def _render_children(
        self,
        children: Sequence[docutils.nodes.Element],
        prefix: Optional[_Token] = None,
        suffix: Optional[_Token] = None,
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        """
        Render the children and concatenate the resulting tokens.

        If ``prefix`` token is defined, it will be prepended to the result.

        If ``suffix`` token is defined, it will be appended to the result
        """
        result = []  # type: List[_Token]

        if prefix is not None:
            result.append(prefix)

        errors = []  # type: List[str]

        for child in children:
            tokens, child_errors = self.transform(child)
            if child_errors is not None:
                errors.extend(child_errors)
                continue

            assert tokens is not None
            result.extend(tokens)

        if len(errors) > 0:
            return None, errors

        if suffix is not None:
            result.append(suffix)

        return result, None

    def transform_paragraph(
        self, element: docutils.nodes.paragraph
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        return self._render_children(children=element.children, prefix=_TokenP())

    def transform_emphasis(
        self, element: docutils.nodes.emphasis
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        return self._render_children(
            children=element.children,
            prefix=_TokenText("<em>"),
            suffix=_TokenText("</em>"),
        )

    def transform_list_item(
        self, element: docutils.nodes.list_item
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        return self._render_children(element.children)

    def transform_bullet_list(
        self, element: docutils.nodes.bullet_list
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        result = [_TokenULOpen()]  # type: List[_Token]

        errors = []  # type: List[str]

        for child in element.children:
            assert isinstance(child, docutils.nodes.list_item), (
                f"Expected a list item in the bullet list, "
                f"but got an instance of {type(child)}: {child}"
            )

            child_tokens, child_errors = self.transform(child)
            if child_errors is not None:
                errors.extend(child_errors)
                continue

            assert child_tokens is not None
            result.append(_TokenLI())
            result.extend(child_tokens)

        result.append(_TokenULClose())

        if len(errors) > 0:
            return None, errors

        return result, None

    def transform_note(
        self, element: docutils.nodes.note
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        return self._render_children(children=element.children, prefix=_TokenP())

    def transform_reference(
        self, element: docutils.nodes.reference
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        text = element.astext()
        escaped_for_href = urllib.parse.quote(element.astext())
        escaped_for_html = html.escape(text)
        return [
            _TokenText(f"<a href='{escaped_for_href}'>{escaped_for_html}</a>")
        ], None

    def transform_field_body(
        self, element: docutils.nodes.field_body
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        return self._render_children(element.children)

    def transform_document(
        self, element: docutils.nodes.document
    ) -> Tuple[Optional[List[_Token]], Optional[List[str]]]:
        return self._render_children(element.children)


def _remove_redundant_p(tokens: List[_Token]) -> List[_Token]:
    """Remove redundant chains of ``<p>`` tokens."""
    result = []  # type: List[_Token]

    # region Remove consecutive ``<p>``
    last_token = None  # type: Optional[_Token]
    for token in tokens:
        if (
            last_token is not None
            and isinstance(last_token, _TokenP)
            and isinstance(token, _TokenP)
        ):
            continue

        result.append(token)
        last_token = token
    # endregion

    # region Remove trailing ``<p>``
    cutoff = 0
    for token in reversed(result):
        if not isinstance(token, _TokenP):
            break

        cutoff += 1

    result = result[: len(result) - cutoff]
    # endregion

    return result


def _strip_prefix_p(tokens: List[_Token]) -> List[_Token]:
    """Strip the ``<p>`` from ``tokens`` if they start with such a token."""
    if len(tokens) == 0:
        return tokens

    if isinstance(tokens[0], _TokenP):
        return tokens[1:]

    return tokens


def _add_prefix_p_if_necessary(tokens: List[_Token]) -> List[_Token]:
    """Add a ``<p>`` at the beginning if the ``tokens`` do not start with a re-flow."""
    if len(tokens) == 0:
        return tokens

    if isinstance(tokens[0], (_TokenP, _TokenULOpen)):
        return tokens

    return [_TokenP()] + tokens


@dataclasses.dataclass
class _TextBlock:
    """Model a text block in the indented output, to be joined with empty strings."""

    text: str


class _RelativeIndention:
    """
    Represent the relative indention.

    Since the indention is *relative*, it can be either positive or negative.
    """

    @require(lambda direction: direction in (-1, 1))
    def __init__(self, direction: int) -> None:
        self.direction = direction

    def __repr__(self) -> str:
        """Generate text representation for easier debugging."""
        return f"{self.__class__.__name__}({self.direction!r})"


_TextDirective = Union[_TextBlock, _RelativeIndention]


class _IndentionMachine:
    """Write the text to a buffer, considering the indention."""

    def __init__(self, indention: str = "  ") -> None:
        self._indention_level = 0
        self._buffer: List[_TextDirective] = []
        self.indention = indention

    def write(self, text: str) -> None:
        """Direct a write of text onto the buffer."""
        self._buffer.append(_TextBlock(text))

    def indent(self) -> None:
        """Direct an indention onto the buffer."""
        self._indention_level += 1
        self._buffer.append(_RelativeIndention(direction=1))

    def dedent(self) -> None:
        """Direct a dedention onto the buffer."""
        self._indention_level -= 1
        if self._indention_level < 0:
            raise ValueError("Unexpected negative indention")

        self._buffer.append(_RelativeIndention(direction=-1))

    def render(self) -> str:
        """Render the buffer as indented text."""
        level = 0
        pending_text_blocks = []  # type: List[_TextBlock]

        writer = io.StringIO()

        for directive in self._buffer:
            if isinstance(directive, _TextBlock):
                pending_text_blocks.append(directive)

            elif isinstance(directive, _RelativeIndention):
                writer.write(
                    textwrap.indent(
                        "".join(block.text for block in pending_text_blocks),
                        self.indention * level,
                    )
                )
                level += directive.direction
                if level < 0:
                    raise AssertionError(
                        "Unexpected indention level below zero; "
                        "expected to catch this in ``dedent()``"
                    )

                pending_text_blocks = []
            else:
                assert_never(directive)

        if len(pending_text_blocks) > 0:
            writer.write(
                textwrap.indent(
                    "".join(block.text for block in pending_text_blocks),
                    self.indention * level,
                )
            )

        return writer.getvalue()


def _render_tokens(tokens: Sequence[_Token]) -> str:
    """Render the output tokens as text."""
    if len(tokens) == 0:
        return ""

    indention_machine = _IndentionMachine()

    for i, token in enumerate(tokens):
        if isinstance(token, _TokenText):
            indention_machine.write(token.content)

        elif isinstance(token, _TokenP):
            if i > 0:
                indention_machine.write("\n\n")

            indention_machine.write("<p>")

        elif isinstance(token, _TokenULOpen):
            if i > 0:
                indention_machine.write("\n\n")

            indention_machine.write("<ul>")
            indention_machine.indent()
        elif isinstance(token, _TokenULClose):
            indention_machine.dedent()
            indention_machine.write("\n</ul>")
        elif isinstance(token, _TokenLI):
            # NOTE (mristin, 2024-03-28):
            # We add a space after ``<li>`` for better readability of the code in, say,
            # a text editor.
            indention_machine.write("\n<li> ")
        else:
            assert_never(token)

    return indention_machine.render()


def documentation_comment(text: Stripped) -> Stripped:
    """Generate a documentation comment out of the text."""
    lines = text.splitlines()
    writer = io.StringIO()
    writer.write("/**\n")
    for line in lines:
        if len(line.strip()) > 0:
            writer.write(f" * {line}\n")
        else:
            writer.write(" *\n")

    writer.write(" */")

    return Stripped(writer.getvalue())


def _post_process_summary(tokens: List[_Token]) -> List[_Token]:
    """Perform the post-processing steps for the summary tokens."""
    return _remove_redundant_p(_strip_prefix_p(tokens))


def _post_process_remark(tokens: List[_Token]) -> List[_Token]:
    """Perform the post-processing steps for the remark tokens."""
    return _remove_redundant_p(_add_prefix_p_if_necessary(tokens))


def _generate_summary_remarks(
    description: intermediate.SummaryRemarksDescription, context: Context
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for a summary-remarks."""
    errors = []  # type: List[str]

    renderer = _ElementRenderer(context=context)

    tokens = []  # type: List[_Token]

    summary_tokens, summary_errors = renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(summary_errors)
    else:
        assert summary_tokens is not None
        tokens.extend(_post_process_summary(summary_tokens))

    for remark in description.remarks:
        remark_tokens, remark_errors = renderer.transform(remark)
        if remark_errors:
            errors.extend(remark_errors)
        else:
            assert remark_tokens is not None
            tokens.extend(_post_process_remark(remark_tokens))

    if len(errors) > 0:
        return None, [
            Error(description.parsed.node, error_message) for error_message in errors
        ]

    return documentation_comment(Stripped(_render_tokens(tokens))), None


def _generate_summary_remarks_constraints(
    description: intermediate.SummaryRemarksConstraintsDescription, context: Context
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for a summary-remarks-constraints."""
    errors = []  # type: List[str]

    renderer = _ElementRenderer(context=context)

    tokens = []  # type: List[_Token]

    summary_tokens, summary_errors = renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(summary_errors)
    else:
        assert summary_tokens is not None
        tokens.extend(_post_process_summary(summary_tokens))

    for remark in description.remarks:
        remark_tokens, remark_errors = renderer.transform(remark)
        if remark_errors:
            errors.extend(remark_errors)
        else:
            assert remark_tokens is not None
            tokens.extend(_post_process_remark(remark_tokens))

    if len(description.constraints_by_identifier) > 0:
        tokens.append(_TokenP())
        tokens.append(_TokenText("Constraints:"))

        tokens.append(_TokenULOpen())

        for constraint_id, constraint in description.constraints_by_identifier.items():
            constraint_tokens, constraint_errors = renderer.transform(constraint)
            if constraint_errors is not None:
                errors.extend(constraint_errors)
            else:
                assert constraint_tokens is not None
                tokens.append(_TokenLI())
                tokens.append(_TokenText(f"Constraint {constraint_id}:\n"))
                tokens.extend(_strip_prefix_p(_remove_redundant_p(constraint_tokens)))

        tokens.append(_TokenULClose())

    if len(errors) > 0:
        return None, [
            Error(description.parsed.node, error_message) for error_message in errors
        ]

    return documentation_comment(Stripped(_render_tokens(tokens))), None


def generate_comment_for_enumeration_literal(
    description: intermediate.DescriptionOfEnumerationLiteral, context: Context
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given enumeration literal."""
    return _generate_summary_remarks(description=description, context=context)


def generate_comment_for_our_type(
    description: intermediate.DescriptionOfOurType, context: Context
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for our type."""
    return _generate_summary_remarks_constraints(
        description=description, context=context
    )


def generate_comment_for_property(
    description: intermediate.DescriptionOfProperty, context: Context
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given property."""
    return _generate_summary_remarks_constraints(
        description=description, context=context
    )


def generate_comment_for_signature(
    description: intermediate.DescriptionOfSignature, context: Context
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """
    Generate the docstring for the given signature.

    A signature, in this context, means a function or a method signature.
    """
    errors = []  # type: List[str]

    renderer = _ElementRenderer(context=context)

    tokens = []  # type: List[_Token]

    summary_tokens, summary_errors = renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(summary_errors)
    else:
        assert summary_tokens is not None
        tokens.extend(_post_process_summary(summary_tokens))

    for remark in description.remarks:
        remark_tokens, remark_errors = renderer.transform(remark)
        if remark_errors:
            errors.extend(remark_errors)
        else:
            assert remark_tokens is not None
            tokens.extend(_post_process_remark(remark_tokens))

    blocks = []  # type: List[Stripped]

    summary_remarks = _render_tokens(tokens)

    if len(summary_remarks) > 0:
        blocks.append(Stripped(summary_remarks))

    param_and_return_blocks = []  # type: List[str]
    for arg_name, arg_description in description.arguments_by_name.items():
        arg_tokens, arg_errors = renderer.transform(arg_description)
        if arg_errors is not None:
            errors.extend(arg_errors)
            continue

        assert arg_tokens is not None

        arg_description = _render_tokens(
            _strip_prefix_p(_remove_redundant_p(arg_tokens))
        )

        param_and_return_blocks.append(Stripped(f"@param {arg_name} {arg_description}"))

    if description.returns is not None:
        returns_tokens, returns_errors = renderer.transform(description.returns)
        if returns_errors is not None:
            errors.extend(returns_errors)
        else:
            assert returns_tokens is not None

            returns_description = _render_tokens(
                _strip_prefix_p(_remove_redundant_p(returns_tokens))
            )

            param_and_return_blocks.append(Stripped(f"@return {returns_description}"))

    if len(errors) > 0:
        return None, [
            Error(description.parsed.node, error_message) for error_message in errors
        ]

    if len(param_and_return_blocks) > 0:
        blocks.append(Stripped("\n".join(param_and_return_blocks)))

    return documentation_comment(Stripped("\n\n".join(blocks))), None



"""Generate TypeScript identifiers based on the identifiers from the meta-model."""
from typing import Union

from icontract import require

import aas_core_codegen.naming
from aas_core_codegen import intermediate
from aas_core_codegen.common import Identifier, assert_never


# fmt: off
@require(
    lambda identifier: identifier[0].isupper(),
    "Enumeration name must start with a capital letter"
)
# fmt: on
def enum_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for an enum based on its meta-model ``identifier``.

    >>> enum_name(Identifier("Something"))
    'Something'

    >>> enum_name(Identifier("URL_to_something"))
    'UrlToSomething'

    >>> enum_name(Identifier("Something_to_URL"))
    'SomethingToUrl'
    """
    return aas_core_codegen.naming.capitalized_camel_case(identifier)


# fmt: off
@require(
    lambda identifier:
    identifier[0].isupper(),
    "Class names must start with a capital letter"
)
# fmt: on
def class_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a class based on its meta-model ``identifier``.

    >>> class_name(Identifier("Something"))
    'Something'

    >>> class_name(Identifier("URL_to_something"))
    'UrlToSomething'

    >>> class_name(Identifier("Something_to_URL"))
    'SomethingToUrl'
    """
    return aas_core_codegen.naming.capitalized_camel_case(identifier)


def name_of(
    something: Union[
        intermediate.Enumeration, intermediate.AbstractClass, intermediate.ConcreteClass
    ]
) -> Identifier:
    """Dispatch the name based on the run-time type of ``something``."""
    if isinstance(something, intermediate.Enumeration):
        return enum_name(something.name)

    elif isinstance(
        something, (intermediate.AbstractClass, intermediate.ConcreteClass)
    ):
        return class_name(something.name)

    else:
        assert_never(something)


def enum_literal_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for an enum literal based on its meta-model ``identifier``.

    >>> enum_literal_name(Identifier("something"))
    'Something'

    >>> enum_literal_name(Identifier("URL_to_something"))
    'UrlToSomething'
    """
    return aas_core_codegen.naming.capitalized_camel_case(identifier)


def constant_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a constant based on its meta-model ``identifier``.

    >>> constant_name(Identifier("something"))
    'SOMETHING'

    >>> constant_name(Identifier("URL_to_something"))
    'URL_TO_SOMETHING'
    """
    parts = identifier.split("_")
    return Identifier("_".join(part.upper() for part in parts))


# fmt: off
@require(
    lambda identifier:
    identifier[0].isupper(),
    "Interface names must start with a capital letter"
)
# fmt: on
def interface_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a class based on its meta-model ``identifier``.

    >>> interface_name(Identifier("Something"))
    'ISomething'

    >>> interface_name(Identifier("URL_to_something"))
    'IUrlToSomething'

    >>> interface_name(Identifier("Something_to_URL"))
    'ISomethingToUrl'
    """
    return Identifier(f"I{aas_core_codegen.naming.capitalized_camel_case(identifier)}")


def property_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a public property based on its meta-model ``identifier``.

    >>> property_name(Identifier("something"))
    'something'

    >>> property_name(Identifier("something_to_URL"))
    'somethingToUrl'

    >>> property_name(Identifier("URL_to_something"))
    'urlToSomething'
    """
    return aas_core_codegen.naming.lower_camel_case(identifier)


def function_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a function from its meta-model ``identifier``.

    >>> function_name(Identifier("do_something"))
    'doSomething'

    >>> function_name(Identifier("do_something_to_URL"))
    'doSomethingToUrl'
    """
    return aas_core_codegen.naming.lower_camel_case(identifier)


def method_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for an instance method based on its meta-model ``identifier``.

    >>> method_name(Identifier("do_something"))
    'doSomething'

    >>> method_name(Identifier("do_something_to_URL"))
    'doSomethingToUrl'
    """
    return aas_core_codegen.naming.lower_camel_case(identifier)


def argument_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for an argument based on its meta-model ``identifier``.

    >>> argument_name(Identifier("something"))
    'something'

    >>> argument_name(Identifier("something_to_URL"))
    'somethingToUrl'
    """
    return aas_core_codegen.naming.lower_camel_case(identifier)


def variable_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a variable based on its meta-model ``identifier``.

    >>> variable_name(Identifier("something"))
    'something'

    >>> variable_name(Identifier("something_to_URL"))
    'somethingToUrl'
    """
    return aas_core_codegen.naming.lower_camel_case(identifier)



"""Encapsulate the entry point to different generators."""
import hashlib
import io
import pathlib
import pickle
import tempfile
import textwrap
import uuid
from typing import Sequence, TextIO, Tuple, Optional, Final

import asttokens
from icontract import require, ensure

import aas_core_codegen
from aas_core_codegen import specific_implementations, intermediate, parse
from aas_core_codegen.common import LinenoColumner


class Context:
    """Represent the context of a code generation."""

    @require(lambda model_path: model_path.exists() and model_path.is_file())
    @require(lambda output_dir: output_dir.exists() and output_dir.is_dir())
    def __init__(
        self,
        model_path: pathlib.Path,
        symbol_table: intermediate.SymbolTable,
        spec_impls: specific_implementations.SpecificImplementations,
        lineno_columner: LinenoColumner,
        output_dir: pathlib.Path,
    ) -> None:
        """Initialize with the given values."""
        self.model_path = model_path
        self.symbol_table = symbol_table
        self.spec_impls = spec_impls
        self.lineno_columner = lineno_columner
        self.output_dir = output_dir


@require(
    lambda errors: all(
        len(error) > 0 and not error.startswith("\n")
        # This is necessary so that we do not have double bullet point.
        and not error.startswith("*") and not error.endswith("\n")
        for error in errors
    )
)
@require(lambda message: not message.endswith(":"))
@require(lambda message: not message.endswith("\n"))
@require(lambda message: not message.startswith("\n") and not message.startswith("*"))
# fmt: on
def write_error_report(message: str, errors: Sequence[str], stderr: TextIO) -> None:
    """
    Write the report (main ``message`` and details as ``errors``) to ``stderr``.

    This method helps us to have a unified way of showing errors.
    """
    stderr.write(f"{message}:\n")
    for error in errors:
        indented = textwrap.indent(error, "  ")
        indented = "* " + indented[2:]
        stderr.write(f"{indented}\n")


class _Cached:
    symbol_table: Final[intermediate.SymbolTable]
    atok: Final[asttokens.ASTTokens]

    def __init__(
        self, symbol_table: intermediate.SymbolTable, atok: asttokens.ASTTokens
    ) -> None:
        self.symbol_table = symbol_table
        self.atok = atok


@require(lambda model_path: model_path.exists() and model_path.is_file())
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def load_model(
    model_path: pathlib.Path, cache_model: bool = False
) -> Tuple[
    Optional[Tuple[intermediate.SymbolTable, asttokens.ASTTokens]], Optional[str]
]:
    """
    Load the given meta-model from the file system and understand it.

    If the ``cache_model`` is set, the symbol table will be stored in the temporary
    directory of your OS keyed on the hash of the model source code. On subsequent
    calls to this function, the model will be loaded from the cache instead of reparsed
    as long as the content of the source code did not change.
    """
    text = model_path.read_text(encoding="utf-8")

    text_hash = hashlib.sha256(text.encode()).hexdigest()
    cache_path = (
        pathlib.Path(tempfile.gettempdir())
        / f"aas-core-codegen-{aas_core_codegen.__version__}"
        / f"model-{text_hash}.pickle"
    )
    if cache_model:
        if cache_path.exists():
            with cache_path.open("rb") as fid:
                cached = pickle.load(fid)
                assert isinstance(cached, _Cached)

                return (cached.symbol_table, cached.atok), None

    atok, parse_exception = parse.source_to_atok(source=text)
    if parse_exception:
        if isinstance(parse_exception, SyntaxError):
            return None, (
                f"Failed to parse the meta-model: "
                f"invalid syntax at line {parse_exception.lineno}"
            )
        else:
            return None, f"Failed to parse the meta-model: {parse_exception}"

    assert atok is not None

    import_errors = parse.check_expected_imports(atok=atok)
    if import_errors:
        writer = io.StringIO()
        write_error_report(
            message="One or more unexpected imports in the meta-model",
            errors=import_errors,
            stderr=writer,
        )
        return None, writer.getvalue()

    lineno_columner = LinenoColumner(atok=atok)

    parsed_symbol_table, error = parse.atok_to_symbol_table(atok=atok)
    if error is not None:
        writer = io.StringIO()
        write_error_report(
            message="Failed to construct the symbol table",
            errors=[lineno_columner.error_message(error)],
            stderr=writer,
        )
        return None, writer.getvalue()

    assert parsed_symbol_table is not None

    ir_symbol_table, error = intermediate.translate(
        parsed_symbol_table=parsed_symbol_table,
        atok=atok,
    )
    if error is not None:
        writer = io.StringIO()
        write_error_report(
            message=(
                "Failed to translate the parsed symbol table "
                "to intermediate symbol table"
            ),
            errors=[lineno_columner.error_message(error)],
            stderr=writer,
        )
        return None, writer.getvalue()

    assert ir_symbol_table is not None

    if cache_model:
        cache_path.parent.mkdir(parents=True, exist_ok=True)

        tmp_path = cache_path.with_suffix(f".{uuid.uuid4()}.tmp")
        try:
            with tmp_path.open("wb") as fid:
                pickle.dump(_Cached(symbol_table=ir_symbol_table, atok=atok), fid)

            tmp_path.rename(cache_path)
        finally:
            tmp_path.unlink(missing_ok=True)

    return (ir_symbol_table, atok), None
